#include <cstdio>
#include <cstddef>
#include <cstdio>
#include <cstdlib>
#include <ctime>
#include <csignal>
#include <exception>
#include <new>
#include <string>
#include <unordered_map>
  #define AUDIO_H_MAINT_ACCESS
#include "audio.h"
  #define SYSTEM_H_MAINT_ACCESS
#include "system.h"
#include "prog.h"
#include "main.h"
#include "utils.h"


namespace Main
{
	static constexpr const char *ver_str = "L engine 8 [1.0]";

	const char *EngineVersion()
	{
		return ver_str;
	}

	const char *GetDefaultMsgTitle()      {return MANIFEST_APPNAME;}
	const char *GetDefaultMsgWarnTitle()  {return MANIFEST_APPNAME " - Warning";}
	const char *GetDefaultMsgErrorTitle() {return MANIFEST_APPNAME " - Error";}

	void Msg(const char *txt, const char *title)
	{
		SDL_ShowSimpleMessageBox(SDL_MESSAGEBOX_INFORMATION, title, txt, 0);
	}
	void Msg_Warn(const char *txt, const char *title)
	{
		SDL_ShowSimpleMessageBox(SDL_MESSAGEBOX_WARNING, title, txt, 0);
	}
	void Msg_Error(const char *txt, const char *title)
	{
		SDL_ShowSimpleMessageBox(SDL_MESSAGEBOX_WARNING, title, txt, 0);
	}

	// "Modal" = freezes the window until closed
	void ModalMsg(const char *txt, const char *title)
	{
		SDL_ShowSimpleMessageBox(SDL_MESSAGEBOX_INFORMATION, title, txt, Window::Handle());
	}
	void ModalMsg_Warn(const char *txt, const char *title)
	{
		SDL_ShowSimpleMessageBox(SDL_MESSAGEBOX_WARNING, title, txt, Window::Handle());
	}
	void ModalMsg_Error(const char *txt, const char *title)
	{
		SDL_ShowSimpleMessageBox(SDL_MESSAGEBOX_WARNING, title, txt, Window::Handle());
	}

	[[noreturn]] void Exit()
	{
		std::exit(0);
	}

	[[noreturn]] void Exit(const char *txt)
	{
		Msg_Error(txt);
		std::exit(0);
	}

	static int r_argc;
	static const char *const *r_argv;

	int argc()
	{
		return r_argc;
	}
	const char *const *argv()
	{
		return r_argv;
	}

	static Main::StateFunc::PtrType current_state = Prog::States::Boot;

	static void Init()
	{
		EXEC_ONCE;

		static constexpr uint32_t sdlinit_flags = SDL_INIT_VIDEO | SDL_INIT_TIMER | SDL_INIT_EVENTS | SDL_INIT_NOPARACHUTE; // Because parachutes are for noobs
		if (SDL_Init(sdlinit_flags))
			Main::Exit(Jo("SDL init failed: `", SDL_GetError(), '`'));
		std::atexit(SDL_Quit);

		System::Init();
		Audio::Init();
		Prog::Init();
	}
}

int main(int argc, char **argv)
{
	EXEC_ONCE;

	auto SignalHandler = [](int sig) -> void
	{
		switch (sig)
		{
		  case SIGINT: case SIGTERM:
		  #ifdef SIGBREAK
		  case SIGBREAK:
		  #endif
			Main::Exit();
			break;

		  case SIGILL:
			Main::Exit("SIGILL: Illegal instruction.");
			break;

		  case SIGFPE:
			Main::Exit("SIGFPE: Floating point error.");
			break;

		  case SIGSEGV:
			Main::Exit("SIGSEGV: Segmentation violation.");
			break;

		  case SIGABRT:
			Main::Exit("SIGABRT: Abnormal termination.");
			break;

		  default:
			Main::Exit("Unknown signal.");
			break;
		}
	};

	std::signal(SIGINT  , SignalHandler);
	std::signal(SIGTERM , SignalHandler);
	std::signal(SIGILL  , SignalHandler);
	std::signal(SIGFPE  , SignalHandler);
	std::signal(SIGSEGV , SignalHandler);
	std::signal(SIGABRT , SignalHandler);
	#ifdef SIGBREAK
	std::signal(SIGBREAK, SignalHandler);
	#endif

	std::set_terminate(
	[]{
		try {throw;}
		catch (const std::exception &e) {Main::Exit(Jo("Standard exception: `", e.what(), "`."));}
		catch (const char *p) {Main::Exit(Jo("Exception: `", p, "`."));}
		catch (...) {Main::Exit(Jo("Unknown exception."));}
	});
	std::set_unexpected(
	[]{
		try {throw;}
		catch (const std::exception &e) {Main::Exit(Jo("Unexpected standard exception: `", e.what(), "`."));}
		catch (const char *p) {Main::Exit(Jo("Unexpected exception: `", p, "`."));}
		catch (...) {Main::Exit(Jo("Unexpected unknown exception."));}
	});

	std::set_new_handler([]{Main::Exit("Required memory is not availble.");});

	Main::r_argc = argc;
	Main::r_argv = argv;

	Main::Init();

	while (Main::current_state) Main::current_state = (Main::StateFunc::PtrType)Main::current_state();

	Main::Exit();

	return 0; // Macro replaced to be `SDL_main` there is no automatic return 0 here!
}