#include "prog.h"

#define ASSETS_DIR "build/assets/"
#define ASSET(x) (ASSETS_DIR x)
#define C_ASSET(x, y) (Utils::CheckIfFileExists(ASSET(x)) ? ASSET(x) : ASSET(x y))

namespace Prog
{
	using MasterLayout = GFX::AttribLayout<M::fvec2, M::fvec2, M::fvec4, M::fvec4, M::fvec4, M::fvec4>;
	Utils::Box<GFX::RenderingQueue<MasterLayout>, Tags::NoAutoCtor> ren_queue;
	Utils::Box<GFX::Texture, Tags::NoAutoCtor> master_tex;
	GFX::CameraController cam_control;
	Utils::Box<GFX::Shader, Tags::NoAutoCtor> master_shader;
	namespace Internal
	{
		unsigned int scale = 0;
		int mx, my;
		int winw, winh;
		M::vec4<M::fvec4> colormat {{1,0,0,0},{0,1,0,0},{0,0,1,0},{0,0,0,1}}; // for 4th vector first 3 vals are not multiplied by src alpha


		unsigned int keys[6];
		std::string keynames[6];

		int world_backgrounds[] {-1,0,-1,-1, 1};     ///      <-------------         WORLDS

		static constexpr int worlds_c = std::extent<decltype(world_backgrounds)>::value;

		struct WorldData
		{
			unsigned int w, h;
			unsigned int *map;
			unsigned int *objmap;
			unsigned int def_id;
		}
		worlds[worlds_c];

		Utils::Box<Audio::Source, Tags::NoAutoCtor> theme_source;
	}

	namespace UniformLocs
	{
		GLint u_mat;
		GLint u_texsz;
		GLint u_scale;
		GLint u_tex;
		GLint u_mastercm;
	}

	void CheckScale()
	{
		unsigned int s = std::min(Window::W() / (MANIFEST_WINDOW_MINW / 2), Window::H() / (MANIFEST_WINDOW_MINH / 2));
		if (!s)
			s = 1;
		if (s != Internal::scale)
		{
			Internal::scale = s;
			glUniform1f(UniformLocs::u_scale, Internal::scale);
		}

		Internal::winw = (Window::W_half() + Internal::scale - 1) / Internal::scale * 2;
		Internal::winh = (Window::H_half() + Internal::scale - 1) / Internal::scale * 2;
	}

	int MouseX() {return Internal::mx;}
	int MouseY() {return Internal::my;}
	int WinW()   {return Internal::winw;}
	int WinH()   {return Internal::winh;}
	int WinWh()  {return Internal::winw/2;}
	int WinHh()  {return Internal::winh/2;}

	void Init()
	{
		GFX::Viewport({0, -Window::H_half()%2}, {Window::W_half()*2, Window::H_half()*2});
		GFX::Blending(1);
		GFX::NormalBlendFunc();
	}

	void Resize()
	{
		GFX::Viewport({0, -Window::H_half()%2}, {Window::W_half()*2, Window::H_half()*2});
		cam_control.ApplyRatioToMatrixes(Window::W_half()*2, Window::H_half()*2);
		cam_control.SetUniView(UniformLocs::u_mat);
		CheckScale();
	}

	namespace Draw
	{
		void Color()
		{
			Internal::colormat = {{1,0,0,0},{0,1,0,0},{0,0,1,0},{0,0,0,1}};
		}
		void Color(M::fvec4 v)
		{
			Internal::colormat = {{v.x,0,0,0},{0,v.y,0,0},{0,0,v.z,0},{0,0,0,v.w}};
		}
		void Color(M::fvec4 sr, M::fvec4 sg, M::fvec4 sb, float sa, M::fvec3 off)
		{
			Internal::colormat = {sr, sg, sb, {off.x,off.y,off.z,sa}};
		}
		void GlobalColor()
		{
			M::vec4<M::fvec4> colormat;
			colormat = {{1,0,0,0},{0,1,0,0},{0,0,1,0},{0,0,0,1}};
			glUniformMatrix4fv(UniformLocs::u_mastercm, 1, 0, (float *)&colormat);
		}
		void GlobalColor(M::fvec4 v)
		{
			M::vec4<M::fvec4> colormat;
			colormat = {{v.x,0,0,0},{0,v.y,0,0},{0,0,v.z,0},{0,0,0,v.w}};
			glUniformMatrix4fv(UniformLocs::u_mastercm, 1, 0, (float *)&colormat);
		}
		void GlobalColor(M::fvec4 sr, M::fvec4 sg, M::fvec4 sb, float sa, M::fvec3 off)
		{
			M::vec4<M::fvec4> colormat;
			colormat = {sr, sg, sb, {off.x,off.y,off.z,sa}};
			glUniformMatrix4fv(UniformLocs::u_mastercm, 1, 0, (float *)&colormat);
		}


		void Rect(M::ivec2 dpos, M::ivec2 dsz, M::ivec2 spos, M::ivec2 ssz)
		{
			dsz += dpos;
			ssz += spos;
			ren_queue->Push4as3x2({M::fvec2{float(dpos.x),float(dpos.y)}, M::fvec2{float(spos.x),float(spos.y)}, Internal::colormat.x, Internal::colormat.y, Internal::colormat.z, Internal::colormat.w},
								  {M::fvec2{float(dsz.x ),float(dpos.y)}, M::fvec2{float(ssz.x ),float(spos.y)}, Internal::colormat.x, Internal::colormat.y, Internal::colormat.z, Internal::colormat.w},
								  {M::fvec2{float(dsz.x ),float(dsz.y )}, M::fvec2{float(ssz.x ),float(ssz.y )}, Internal::colormat.x, Internal::colormat.y, Internal::colormat.z, Internal::colormat.w},
								  {M::fvec2{float(dpos.x),float(dsz.y )}, M::fvec2{float(spos.x),float(ssz.y )}, Internal::colormat.x, Internal::colormat.y, Internal::colormat.z, Internal::colormat.w});
		}
		void Rect(M::ivec2 dpos, M::ivec2 dsz, M::ivec2 spos)
		{
			M::ivec2 ssz = spos + dsz;
			dsz += dpos;
			ren_queue->Push4as3x2({M::fvec2{float(dpos.x),float(dpos.y)}, M::fvec2{float(spos.x),float(spos.y)}, Internal::colormat.x, Internal::colormat.y, Internal::colormat.z, Internal::colormat.w},
								  {M::fvec2{float(dsz.x ),float(dpos.y)}, M::fvec2{float(ssz.x ),float(spos.y)}, Internal::colormat.x, Internal::colormat.y, Internal::colormat.z, Internal::colormat.w},
								  {M::fvec2{float(dsz.x ),float(dsz.y )}, M::fvec2{float(ssz.x ),float(ssz.y )}, Internal::colormat.x, Internal::colormat.y, Internal::colormat.z, Internal::colormat.w},
								  {M::fvec2{float(dpos.x),float(dsz.y )}, M::fvec2{float(spos.x),float(ssz.y )}, Internal::colormat.x, Internal::colormat.y, Internal::colormat.z, Internal::colormat.w});
		}

		void RectFill(M::ivec2 dpos, M::ivec2 dsz, M::ivec2 spos)
		{
			dsz += dpos;
			ren_queue->Push4as3x2({M::fvec2{float(dpos.x),float(dpos.y)}, spos + 0.5, Internal::colormat.x, Internal::colormat.y, Internal::colormat.z, Internal::colormat.w},
								  {M::fvec2{float(dsz.x ),float(dpos.y)}, spos + 0.5, Internal::colormat.x, Internal::colormat.y, Internal::colormat.z, Internal::colormat.w},
								  {M::fvec2{float(dsz.x ),float(dsz.y )}, spos + 0.5, Internal::colormat.x, Internal::colormat.y, Internal::colormat.z, Internal::colormat.w},
								  {M::fvec2{float(dpos.x),float(dsz.y )}, spos + 0.5, Internal::colormat.x, Internal::colormat.y, Internal::colormat.z, Internal::colormat.w});
		}
		void RectFill_LineH(M::ivec2 dpos, M::ivec2 dsz, M::ivec2 spos, int len)
		{
			dsz += dpos;
			ren_queue->Push4as3x2({M::fvec2{float(dpos.x),float(dpos.y)}, {float(spos.x      ), spos.y + 0.5f}, Internal::colormat.x, Internal::colormat.y, Internal::colormat.z, Internal::colormat.w},
								  {M::fvec2{float(dsz.x ),float(dpos.y)}, {float(spos.x + len), spos.y + 0.5f}, Internal::colormat.x, Internal::colormat.y, Internal::colormat.z, Internal::colormat.w},
								  {M::fvec2{float(dsz.x ),float(dsz.y )}, {float(spos.x + len), spos.y + 0.5f}, Internal::colormat.x, Internal::colormat.y, Internal::colormat.z, Internal::colormat.w},
								  {M::fvec2{float(dpos.x),float(dsz.y )}, {float(spos.x      ), spos.y + 0.5f}, Internal::colormat.x, Internal::colormat.y, Internal::colormat.z, Internal::colormat.w});
		}
		void RectFill_LineV(M::ivec2 dpos, M::ivec2 dsz, M::ivec2 spos, int len)
		{
			dsz += dpos;
			ren_queue->Push4as3x2({M::fvec2{float(dpos.x),float(dpos.y)}, {spos.x + 0.5f, float(spos.y      )}, Internal::colormat.x, Internal::colormat.y, Internal::colormat.z, Internal::colormat.w},
								  {M::fvec2{float(dsz.x ),float(dpos.y)}, {spos.x + 0.5f, float(spos.y      )}, Internal::colormat.x, Internal::colormat.y, Internal::colormat.z, Internal::colormat.w},
								  {M::fvec2{float(dsz.x ),float(dsz.y )}, {spos.x + 0.5f, float(spos.y + len)}, Internal::colormat.x, Internal::colormat.y, Internal::colormat.z, Internal::colormat.w},
								  {M::fvec2{float(dpos.x),float(dsz.y )}, {spos.x + 0.5f, float(spos.y + len)}, Internal::colormat.x, Internal::colormat.y, Internal::colormat.z, Internal::colormat.w});
		}

		void RectRot(M::fvec2 dpos, M::ivec2 dsz, M::ivec2 spos, M::ivec2 ssz, float angle, M::fvec2 center)
		{
			M::fvec2 tc = center / ssz * dsz;
			M::fvec2 b1 = -tc, b3 = dsz - tc, b2{b3.x, b1.y}, b4{b1.x, b3.y};
			M::fvec2 sx{std::cos(angle), std::sin(angle)}, sy{-sx.y, sx.x};
			b1 = b1.x * sx + b1.y * sy;
			b2 = b2.x * sx + b2.y * sy;
			b3 = b3.x * sx + b3.y * sy;
			b4 = b4.x * sx + b4.y * sy;
			b1 += dpos;
			b2 += dpos;
			b3 += dpos;
			b4 += dpos;
			ssz += spos;
			ren_queue->Push4as3x2({b1, M::fvec2{float(spos.x),float(spos.y)}, Internal::colormat.x, Internal::colormat.y, Internal::colormat.z, Internal::colormat.w},
								  {b2, M::fvec2{float(ssz.x ),float(spos.y)}, Internal::colormat.x, Internal::colormat.y, Internal::colormat.z, Internal::colormat.w},
								  {b3, M::fvec2{float(ssz.x ),float(ssz.y )}, Internal::colormat.x, Internal::colormat.y, Internal::colormat.z, Internal::colormat.w},
								  {b4, M::fvec2{float(spos.x),float(ssz.y )}, Internal::colormat.x, Internal::colormat.y, Internal::colormat.z, Internal::colormat.w});
		}
		void RectRot(M::fvec2 dpos, M::ivec2 dsz, M::ivec2 spos, float angle, M::fvec2 center)
		{
			M::ivec2 ssz = dsz;
			M::fvec2 tc = center / ssz * dsz;
			M::fvec2 b1 = -tc, b3 = dsz - tc, b2{b3.x, b1.y}, b4{b1.x, b3.y};
			M::fvec2 sx{std::cos(angle), std::sin(angle)}, sy{-sx.y, sx.x};
			b1 = b1.x * sx + b1.y * sy;
			b2 = b2.x * sx + b2.y * sy;
			b3 = b3.x * sx + b3.y * sy;
			b4 = b4.x * sx + b4.y * sy;
			b1 += dpos;
			b2 += dpos;
			b3 += dpos;
			b4 += dpos;
			ssz += spos;
			ren_queue->Push4as3x2({b1, M::fvec2{float(spos.x),float(spos.y)}, Internal::colormat.x, Internal::colormat.y, Internal::colormat.z, Internal::colormat.w},
								  {b2, M::fvec2{float(ssz.x ),float(spos.y)}, Internal::colormat.x, Internal::colormat.y, Internal::colormat.z, Internal::colormat.w},
								  {b3, M::fvec2{float(ssz.x ),float(ssz.y )}, Internal::colormat.x, Internal::colormat.y, Internal::colormat.z, Internal::colormat.w},
								  {b4, M::fvec2{float(spos.x),float(ssz.y )}, Internal::colormat.x, Internal::colormat.y, Internal::colormat.z, Internal::colormat.w});
		}

		void Text(M::ivec2 dpos, const char *txt, int align = -1)
		{
			M::ivec2 cpos = dpos;
			auto Shift = [&](const char *p)
			{
				switch (align)
				{
				  default:
				  case -1:
					break;
				  case 0:
				  case 1:
					while (*p && *p != '\n')
					{
						cpos.x = cpos.x - 3 - align * 3;
						p++;
					}
				}
			};
			Shift(txt);
			while (*txt)
			{
				if (*txt != '\n')
				{
					Rect(cpos, {6,12}, {*(unsigned char *)txt % 16 * 6, *(unsigned char *)txt / 16 * 12});
					cpos.x += 6;
				}
				else
				{
					cpos.x = dpos.x;
					cpos.y += 12;
					Shift(txt + 1);
				}
				txt++;
			}
		}
		void Text(M::ivec2 dpos, const char *txt, M::fvec4 color, int align = -1, M::ivec2 shoff = {0,0}, M::fvec4 shcolor = {0,0,0,1})
		{
			M::vec4<M::fvec4> back = Internal::colormat;

			if (shoff)
			{
				Color(shcolor);
				Text(dpos + shoff, txt, align);
			}
			Color(color);
			Text(dpos, txt, align);

			Internal::colormat = back;
		}
	}

	namespace Sound   ///       <------------    SOUND
	{
		namespace Buf
		{
			using Type = Utils::Box<Audio::Buffer, Tags::NoAutoCtor>;
			Type err;
			Type ok;
			Type dia;
			Type death;
			Type save;
			Type load;
			Type teleport;
			Type pistol;
			Type mobdeath;
			Type rot;
			Type tshot;
			Type field;
			Type endtheme;

			Type maintheme;
		}
	}



	struct DynObjData
	{
		unsigned int id;
		bool dead;
		M::ivec2 pos;
		M::fvec2 vel;
		bool facing_right;
		int anim_frame;
		unsigned int walk_time;
		M::ivec2 home_loc;
		std::vector<M::ivec2> data;
	};

	struct PlayerData
	{
		M::ivec2 pos;
		M::fvec2 vel;
		unsigned int walk_time;
		int anim_frame;
		M::ivec2 tex_off;
		int aim_y;
		int current_gun;
		int time_since_shot;
		bool facing_right;
		bool dead;
		int death_timer;
	};

	struct WorldData
	{
		unsigned long long time;
		PlayerData player;
		unsigned int worldnum;
		M::ivec2 campos;

		std::list<DynObjData> dynamic_objs;

		float fade_time;
		float fade_step;
		M::fvec4 fade_color;

		const char *hint_txt;
		int hint_timer;
		M::fvec4 hint_color;

		bool dialogue;
		std::string dia_text;
		const char *dia_rem_text;
		std::list<std::string> dia_rem_pages;
		unsigned int dia_icon;
		M::fvec4 dia_color;
		int dia_timer;
		bool dia_self_speech;

		std::string msg_text;

		bool next_world_req;
		int next_world_timer;
		M::ivec2 next_world_loc;
		bool next_world_face_right;
		M::fvec4 next_world_fade_color;

		bool talked_with_center;
		int aliens_killed;
		bool talked_with_center_2;

		bool wowmsg_0;
		bool wowmsg_1;
		bool wowmsg_2;

		bool girlmsg_0;
		bool girlmsg_1;
		bool girlmsg_2;

		bool ending;
		int ending_timer;
	};



	WorldData saved_world;

	void DefaultTileDraw(M::ivec2 pos, unsigned int world, M::ivec2 mappos);

	void DefaultObjDraw(M::ivec2 pos, unsigned int world, M::ivec2 mappos);
	void TileObjDraw(M::ivec2 pos, unsigned int world, M::ivec2 mappos);

	void DefaultDynObjDraw(M::ivec2 pos, DynObjData &o);
	void DynObjDraw_Bullet(M::ivec2 pos, DynObjData &o);

	namespace Config                                                            ///     <--------------- TILES ARE HERE
	{
		struct TileTypeData
		{
			M::ucvec4 color_code;
			bool visible;
			M::ivec2 texpos;
			bool solid;
			bool merge;
			std::vector<int> merge_with; // always merges with itself
			void (*draw_func)(M::ivec2 pos, unsigned int world, M::ivec2 mappos);
		};

		const TileTypeData list[]
		{ //color               visible  texpos         solid  merge  m.with      draw func
			{{255,255,255,  0}, 0,       {    0,   0},  0,     0,     {},           DefaultTileDraw}, // 0 air
			{{255,255,255,  0}, 0,       {    0,   0},  1,     0,     {},           DefaultTileDraw}, // 1 solid air
			{{127,127,127,255}, 1,       { 0*16, 320},  1,     1,     {},           DefaultTileDraw}, // 2 white tiles
			{{255,255,255,255}, 1,       { 1*16, 320},  0,     1,     {5,6},        DefaultTileDraw}, // 3 white walls
			{{ 63, 63, 63,255}, 1,       { 2*16, 320},  1,     1,     {},           DefaultTileDraw}, // 4 black blocks
			{{191,191,191,255}, 1,       { 3*16, 320},  0,     1,     {3,6},        DefaultTileDraw}, // 5 wall chains
			{{255,255,127,255}, 1,       { 4*16, 320},  0,     1,     {3,5},        DefaultTileDraw}, // 6 alt white walls
			{{  0,127,255,255}, 1,       { 6*16, 320},  1,     1,     {2,4,9},      DefaultTileDraw}, // 7 planet surface
			{{  0, 63,127,255}, 1,       { 7*16, 320},  0,     1,     {3,5,6,7,10}, DefaultTileDraw}, // 8 planet surface walls
			{{  0,  0,255,255}, 1,       { 9*16, 320},  1,     1,     {},           DefaultTileDraw}, // 9 planet bricks
			{{  0,  0,127,255}, 1,       {10*16, 320},  0,     1,     {9},          DefaultTileDraw}, //10 planet brick walls
			{{  0,255,  0,255}, 1,       {11*16, 320},  1,     1,     {},           DefaultTileDraw}, //11 alt planet bricks
			{{  0,127,  0,255}, 1,       {12*16, 320},  0,     1,     {11},         DefaultTileDraw}, //12 alt planet brick walls
			{{  0,  0,  0,255}, 1,       {13*16, 320},  0,     1,     {},           DefaultTileDraw}, //13 black block walls
			{{255,  0,  0,255}, 1,       {14*16, 320},  1,     1,     {},           DefaultTileDraw}, //14 alt 2 white tile
			{{127,  0,  0,255}, 1,       {15*16, 320},  0,     1,     {},           DefaultTileDraw}, //15 alt 2 white walls
			{{255,255,  0,255}, 1,       {16*16, 320},  0,     1,     {},           DefaultTileDraw}, //16 fake sky walls
			{{255,  0,255,255}, 1,       {17*16, 320},  1,     1,     {},           DefaultTileDraw}, //17 dirt
			{{127,  0,127,255}, 1,       {17*16, 320},  0,     1,     {17},         DefaultTileDraw}, //18 dirt walls
		};

		struct ObjTypeData
		{
			M::ucvec4 color_code;
			bool visible;
			M::ivec2 texpos;
			M::ivec2 texsize;
			M::ivec2 texoff;
			void (*draw_func)(M::ivec2 pos, unsigned int world, M::ivec2 mappos);
			std::vector<unsigned int> merge_with;
			unsigned int spawn_dyn;
			bool spawn_once;
			bool respawn_if_dead;
			bool spawn_right;
			M::ivec2 spawn_off;
		};

		const ObjTypeData objlist[]
		{ //color               visible  texpos         texsize   texoff      draw func       merge              spawn  once  resp dead  sp r. spawn off
			{{255,255,255,255}, 0,       {    0,   0},  { 0, 0},  {  0,  0},  DefaultObjDraw, {},                0,     0,    0,         0,    {0,0}}, // 0 nothing
			{{255,255,  0,255}, 1,       {    0, 400},  {48,32},  {-16,  0},  DefaultObjDraw, {},                0,     0,    0,         0,    {0,0}}, // 1 lamp
			{{  0,  0,255,255}, 0,       {    0,   0},  { 0, 0},  {  0,  0},  DefaultObjDraw, {},                2,     0,    1,         0,    {8,0}}, // 2 base girl spawner
			{{  0,255,  0,255}, 0,       {    0,   0},  { 0, 0},  {  0,  0},  DefaultObjDraw, {},                1,     0,    1,         1,    {8,0}}, // 3 checkpoint spawner
			{{255,  0,  0,255}, 1,       { 16*5, 320},  {16,16},  {  0,  0},  TileObjDraw   , {1,2,4,7,9,11,14}, 0,     0,    0,         0,    {0,0}}, // 4 spikes
			{{255,  0,255,255}, 1,       {    0,   0},  { 0, 0},  {  0,  0},  DefaultObjDraw, {},                3,     0,    1,         1,    {8,0}}, // 5 teleporter spawner
			{{127,  0,127,255}, 1,       {    0,   0},  { 0, 0},  {  0,  0},  DefaultObjDraw, {},                4,     0,    1,         1,    {8,0}}, // 6 useless teleporter spawner
			{{  0,  0,  0,255}, 1,       {    0,   0},  { 0, 0},  {  0,  0},  DefaultObjDraw, {},                7,     1,    0,         0,    {8,0}}, // 7 alien spawner
			{{255,127,  0,255}, 1,       { 16*8, 320},  {16,16},  {  0,  0},  TileObjDraw   , {1,2,4,7,9,11,14}, 0,     0,    0,         0,    {0,0}}, // 8 planet spikes
			{{  0,255,255,255}, 1,       {  144, 112},  {32,32},  {  0,  0},  DefaultObjDraw, {},                0,     0,    0,         0,    {0,0}}, // 9 blood 1
			{{  0,127,255,255}, 1,       {  176, 112},  {32,32},  {  0,  0},  DefaultObjDraw, {},                0,     0,    0,         0,    {0,0}}, //10 blood 2
			{{ 63, 63, 63,255}, 1,       {    0,   0},  { 0, 0},  {  0,  0},  DefaultObjDraw, {},                9,     0,    0,         0,    {8,0}}, //11 turret spawner
			{{  0,  0,127,255}, 1,       {    0,   0},  { 0, 0},  {  0,  0},  DefaultObjDraw, {},               10,     0,    0,         1,    {8,8}}, //12 field spawner
			{{127,127,127,255}, 1,       {    0,   0},  { 0, 0},  {  0,  0},  DefaultObjDraw, {},               11,     1,    0,         1,    {8,0}}, //13 base girl 2 spawner
			{{191,191,191,255}, 1,       {  208, 112},  {16,32},  {  0,-16},  DefaultObjDraw, {},                0,     0,    0,         0,    {0,0}}, //14 grass
		};

		struct DynObjTypeData
		{
			bool visible;
			bool phys;
			bool persistent;
			M::ivec2 texpos;
			M::ivec2 texsize;
			M::ivec2 texoff;
			void (*draw_func)(M::ivec2 pos, DynObjData &o);
			std::vector<M::ivec2> hbox;
		};

		const DynObjTypeData dynobjlist[]
		{ //vis  phys  pers  texpos     texsize  texoff     draw func          hitbox
			{0,  0,    0,    {  0,  0}, { 0, 0}, {  0,  0}, DefaultDynObjDraw, {}}, // 0 (marker)
			{1,  0,    0,    {160,  0}, {32,32}, {-16,-16}, DefaultDynObjDraw, {}}, // 1 checkpoint
			{1,  1,    1,    {288, 48}, {24,32}, {-12,-16}, DefaultDynObjDraw, {{-3,-13},{ 2,-13},{-3,  2},{ 2,  2},{-3, 15},{ 2, 15}}}, // 2 base girl
			{1,  0,    0,    {320,192}, {32,48}, {-16,-32}, DefaultDynObjDraw, {}}, // 3 teleporter
			{1,  0,    0,    {320,192}, {32,48}, {-16,-32}, DefaultDynObjDraw, {}}, // 4 useless teleporter
			{1,  0,    0,    {  0,  0}, { 0, 0}, {  0,  0}, DynObjDraw_Bullet, {}}, // 5 player bullet
			{0,  0,    1,    {  0,  0}, { 0, 0}, {  0,  0}, DefaultDynObjDraw, {}}, // 6 (persistent marker)
			{1,  1,    1,    {312, 48}, {24,32}, {-12,-16}, DefaultDynObjDraw, {}}, // 7 alien
			{1,  1,    1,    {336, 48}, {24,32}, {-12,-16}, DefaultDynObjDraw, {{-3,-13},{ 2,-13},{-3,  2},{ 2,  2},{-3, 15},{ 2, 15}}}, // 8 oldman
			{1,  1,    0,    {360, 48}, {24,32}, {-12,-16}, DefaultDynObjDraw, {{-3,-13},{ 2,-13},{-3,  2},{ 2,  2},{-3, 15},{ 2, 15}}}, // 9 turret
			{1,  0,    0,    {208, 64}, {16,16}, { -8, -8}, DefaultDynObjDraw, {}}, //10 field
			{1,  0,    1,    {384, 48}, {32,32}, {-16,-16}, DefaultDynObjDraw, {{-3,-13},{ 2,-13},{-3,  2},{ 2,  2},{-3, 15},{ 2, 15}}}, //11 base girl 2
		};

		M::ivec2 player_hitbox[] {{-3,-14},
								  { 2,-14},
								  {-3,  2},
								  { 2,  2},
								  {-3, 15},
								  { 2, 15}};
	}

	void DefaultTileDraw(M::ivec2 pos, unsigned int world, M::ivec2 mappos)
	{
		unsigned int id = Game::TileAtCoords(world, mappos);
		const auto &p = Config::list[id];

		if (p.merge == 0)
		{
			Draw::Rect(pos, {16,16}, p.texpos);
		}
		else
		{
			bool flags[8]{}; // right, down-right, down, ... , up-right
			unsigned int near[8] {Game::TileAtCoords(world, {mappos.x+1, mappos.y}),   // 0
								  Game::TileAtCoords(world, {mappos.x+1, mappos.y+1}), // 1
								  Game::TileAtCoords(world, {mappos.x  , mappos.y+1}), // 2
								  Game::TileAtCoords(world, {mappos.x-1, mappos.y+1}), // 3
								  Game::TileAtCoords(world, {mappos.x-1, mappos.y  }), // 4
								  Game::TileAtCoords(world, {mappos.x-1, mappos.y-1}), // 5
								  Game::TileAtCoords(world, {mappos.x  , mappos.y-1}), // 6
								  Game::TileAtCoords(world, {mappos.x+1, mappos.y-1})};// 7

			for (int i = 0; i < 8; i++)
			{
				if (id == near[i])
				{
					flags[i] = 1;
				}
			}
			for (unsigned int it : p.merge_with)
			{
				for (int i = 0; i < 8; i++)
				{
					if (it == near[i])
					{
						flags[i] = 1;
					}
				}
			}

			auto DoMerge = [&](int x, int y) -> bool
			{
				switch (x)
				{
				  case 1:
					switch (y)
					{
					  case 1:
						return flags[1];
					  case 0:
						return flags[0];
					  case -1:
						return flags[7];
					}
					break;
				  case 0:
					switch (y)
					{
					  case 1:
						return flags[2];
					  case -1:
						return flags[6];
					}
					break;
				  case -1:
					switch (y)
					{
					  case 1:
						return flags[3];
					  case 0:
						return flags[4];
					  case -1:
						return flags[5];
					}
					break;
				}
				return 0;
			};

			for (int y : {-1, 1})
			{
				for (int x : {-1, 1})
				{
					int num;
					if (DoMerge(x,0) && DoMerge(0,y) && DoMerge(x,y))
					{
						num = 4;
					}
					else
					{
						if (DoMerge(x,0) && DoMerge(0,y))
							num = 3;
						else if (DoMerge(0,y))
							num = 2;
						else if (DoMerge(x,0))
							num = 1;
						else
							num = 0;
					}
					Draw::Rect({pos.x+4+4*x,pos.y+4+4*y},{8,8},p.texpos + M::ivec2{4+4*x,4+4*y+num*16});
				}
			}
		}
	}
	void TileObjDraw(M::ivec2 pos, unsigned int world, M::ivec2 mappos)
	{
		unsigned int id = Game::ObjAtCoords(world, mappos);
		const auto &p = Config::objlist[id];

		bool flags[8]{}; // right, down-right, down, ... , up-right
		unsigned int near[8] {Game::TileAtCoords(world, {mappos.x+1, mappos.y}),   // 0
							  Game::TileAtCoords(world, {mappos.x+1, mappos.y+1}), // 1
							  Game::TileAtCoords(world, {mappos.x  , mappos.y+1}), // 2
							  Game::TileAtCoords(world, {mappos.x-1, mappos.y+1}), // 3
							  Game::TileAtCoords(world, {mappos.x-1, mappos.y  }), // 4
							  Game::TileAtCoords(world, {mappos.x-1, mappos.y-1}), // 5
							  Game::TileAtCoords(world, {mappos.x  , mappos.y-1}), // 6
							  Game::TileAtCoords(world, {mappos.x+1, mappos.y-1})};// 7

		unsigned int onear[8]{Game::ObjAtCoords(world, {mappos.x+1, mappos.y}),   // 0
							  Game::ObjAtCoords(world, {mappos.x+1, mappos.y+1}), // 1
							  Game::ObjAtCoords(world, {mappos.x  , mappos.y+1}), // 2
							  Game::ObjAtCoords(world, {mappos.x-1, mappos.y+1}), // 3
							  Game::ObjAtCoords(world, {mappos.x-1, mappos.y  }), // 4
							  Game::ObjAtCoords(world, {mappos.x-1, mappos.y-1}), // 5
							  Game::ObjAtCoords(world, {mappos.x  , mappos.y-1}), // 6
							  Game::ObjAtCoords(world, {mappos.x+1, mappos.y-1})};// 7

		for (int i = 0; i < 8; i++)
		{
			if (id == onear[i])
			{
				flags[i] = 1;
			}
		}
		for (unsigned int it : p.merge_with)
		{
			for (int i = 0; i < 8; i++)
			{
				if (it == near[i])
				{
					flags[i] = 1;
				}
			}
		}

		auto DoMerge = [&](int x, int y) -> bool
		{
			switch (x)
			{
			  case 1:
				switch (y)
				{
				  case 1:
					return flags[1];
				  case 0:
					return flags[0];
				  case -1:
					return flags[7];
				}
				break;
			  case 0:
				switch (y)
				{
				  case 1:
					return flags[2];
				  case -1:
					return flags[6];
				}
				break;
			  case -1:
				switch (y)
				{
				  case 1:
					return flags[3];
				  case 0:
					return flags[4];
				  case -1:
					return flags[5];
				}
				break;
			}
			return 0;
		};

		for (int y : {-1, 1})
		{
			for (int x : {-1, 1})
			{
				int num;
				if (DoMerge(x,0) && DoMerge(0,y) && DoMerge(x,y))
				{
					num = 4;
				}
				else
				{
					if (DoMerge(x,0) && DoMerge(0,y))
						num = 3;
					else if (DoMerge(0,y))
						num = 2;
					else if (DoMerge(x,0))
						num = 1;
					else
						num = 0;
				}
				Draw::Rect({pos.x+4+4*x,pos.y+4+4*y},{8,8},p.texpos + M::ivec2{4+4*x,4+4*y+num*16});
			}
		}
	}

	void DefaultObjDraw(M::ivec2 pos, unsigned int world, M::ivec2 mappos)
	{
		const auto &ref = Config::objlist[Game::ObjAtCoords(world, mappos)];
		Draw::Rect(pos+ref.texoff, ref.texsize, ref.texpos);
	}

	void DefaultDynObjDraw(M::ivec2 pos, DynObjData &o)
	{
		const auto &c = Config::dynobjlist[o.id];
		int dir = o.facing_right ? 1 : -1;
		Draw::Rect(pos + M::ivec2{c.texoff.x * dir, c.texoff.y}, M::ivec2{c.texsize.x*dir, c.texsize.y}, c.texpos, c.texsize);
	}

	void DynObjDraw_Bullet(M::ivec2 pos, DynObjData &o)
	{
		if (o.vel.y == 0)
			Draw::RectFill({pos.x-4,pos.y},{9,1},{2,0});
		else
			Draw::RectFill({pos.x,pos.y-4},{1,9},{2,0});
	}


	namespace Game
	{
		static constexpr const char *keyconfig_filename = "controls_config.txt";

		void UpdateKeyName(int i)
		{
			const char *tmp = SDL_GetScancodeName((SDL_Scancode)Internal::keys[i]);
			if (!tmp || !*tmp)
			{
				Internal::keynames[i] = "<unknown>";
			}
			else
			{
				Internal::keynames[i] = tmp;
				if (std::strlen(tmp) > 15)
					Internal::keynames[i].resize(15);
			}
		}
		void UpdateKeyNames()
		{
			for (int i = 0; i < 6; i++)
				UpdateKeyName(i);
		}
		void SaveKeyConfig()
		{
			std::ofstream out(keyconfig_filename);
			if (!out)
				return;
			for (unsigned int it : Internal::keys)
				out << it << ' ';
		}
		const char *KeyName_Left () {return Internal::keynames[0].c_str();}
		const char *KeyName_Right() {return Internal::keynames[1].c_str();}
		const char *KeyName_Up   () {return Internal::keynames[2].c_str();}
		const char *KeyName_Down () {return Internal::keynames[3].c_str();}
		const char *KeyName_Jump () {return Internal::keynames[4].c_str();}
		const char *KeyName_Shoot() {return Internal::keynames[5].c_str();}
		unsigned int KeyLeft () {return Internal::keys[0];}
		unsigned int KeyRight() {return Internal::keys[1];}
		unsigned int KeyUp   () {return Internal::keys[2];}
		unsigned int KeyDown () {return Internal::keys[3];}
		unsigned int KeyJump () {return Internal::keys[4];}
		unsigned int KeyShoot() {return Internal::keys[5];}

		void DrawFlatBack()
		{
			int w = (WinW() + 31) / 32;
			int h = (WinH() + 31) / 32;
			for (int i = 0; i < h; i++)
			{
				for (int j = 0; j < w; j++)
				{
					Draw::Rect({j*32-WinWh(),i*32-WinHh()}, {32,32}, {96,0});
				}
			}
		}

		void DrawBackground(int id)
		{
			M::ivec2 pos{704,id * 240};
			Draw::Rect({-160,-120}, {320,240}, pos);
			Draw::RectFill({-WinWh(), -WinHh()},{WinWh()-160,WinHh()-120}, pos);
			Draw::RectFill({160, -WinHh()},{WinWh()-160,WinHh()-120}, {pos.x+320-1, pos.y});
			Draw::RectFill({160, 120},{WinWh()-160,WinHh()-120}, {pos.x+320-1, pos.y+240-1});
			Draw::RectFill({-WinWh(), 120},{WinWh()-160,WinHh()-120}, {pos.x, pos.y+240-1});
			Draw::RectFill_LineH({-160,-WinHh()},{320,WinHh()-120}, pos, 320);
			Draw::RectFill_LineH({-160,120},{320,WinHh()-120}, {pos.x, pos.y+240-1}, 320);
			Draw::RectFill_LineV({-WinWh(),-120},{WinWh()-160,240}, pos, 240);
			Draw::RectFill_LineV({160,-120},{WinWh()-160,240}, {pos.x+320-1, pos.y}, 240);
		}

		void DrawHint(WorldData &world, const char *txt, M::fvec4 color = {1, 0.5, 0, 1})
		{
			static std::string buf;
			buf = txt;
			world.hint_txt = buf.c_str();
			world.hint_timer += 2;
			world.hint_color = color;
		}

		void DrawMessage(WorldData &world, const char *txt)
		{
			world.msg_text += txt;
			world.msg_text += "\n";
		}

		void Fade(WorldData &world, float time, float add_time, M::fvec4 color)
		{
			world.fade_step = 1.0f / (time * 60);
			world.fade_time = 1 + add_time*(1.0f / time);
			world.fade_color = color;
		}

		void ForceRender()
		{
			ren_queue->Send();
			ren_queue->Draw();
			ren_queue->Clear();
		}

		void StartDialogue(WorldData &w, std::list<const char *> pages, unsigned int icon, M::fvec4 color)
		{
			if (pages.size() == 0)
				return;
			if (w.dialogue)
				return;
			w.dialogue = 1;
			w.dia_rem_pages.clear();
			if (**pages.begin() == '\0')
			{
				pages.erase(pages.begin());
				w.dia_self_speech = 1;
			}
			else
			{
				w.dia_self_speech = 0;
			}
			for (const char *it : pages)
				w.dia_rem_pages.emplace_back(it);
			w.dia_rem_text = w.dia_rem_pages.begin()->c_str();
			w.dia_text = "";
			w.dia_icon = icon;
			w.dia_color = color;
			w.dia_timer = 30;
		}

		void PlayErrorSound()
		{
			Audio::Source src(*Sound::Buf::err, Audio::Source::RepeatMode::single_keep);
			src.Play();
		}
		void PlayOkSound()
		{
			Audio::Source src(*Sound::Buf::ok, Audio::Source::RepeatMode::single_keep);
			src.Play();
		}
		void PlayDiaSound()
		{
			Audio::Source src(*Sound::Buf::dia, Audio::Source::RepeatMode::single_keep);
			src.Play();
		}
		void PlayDeathSound()
		{
			Audio::Source src(*Sound::Buf::death, Audio::Source::RepeatMode::single_keep);
			src.Play();
		}
		void PlaySaveSound()
		{
			Audio::Source src(*Sound::Buf::save, Audio::Source::RepeatMode::single_keep);
			src.Play();
		}
		void PlayLoadSound()
		{
			Audio::Source src(*Sound::Buf::load, Audio::Source::RepeatMode::single_keep);
			src.Play();
		}
		void PlayTeleportSound()
		{
			Audio::Source src(*Sound::Buf::teleport, Audio::Source::RepeatMode::single_keep);
			src.Play();
		}
		void PlayPistolSound()
		{
			Audio::Source src(*Sound::Buf::pistol, Audio::Source::RepeatMode::single_keep);
			src.Play();
		}
		void PlayMobdeathSound()
		{
			Audio::Source src(*Sound::Buf::mobdeath, Audio::Source::RepeatMode::single_keep);
			src.Play();
		}
		void PlayRotSound()
		{
			Audio::Source src(*Sound::Buf::rot, Audio::Source::RepeatMode::single_keep);
			src.Play();
		}
		void PlayTshotSound()
		{
			Audio::Source src(*Sound::Buf::tshot, Audio::Source::RepeatMode::single_keep);
			src.Play();
		}
		void PlayFieldSound()
		{
			Audio::Source src(*Sound::Buf::field, Audio::Source::RepeatMode::single_keep);
			src.Play();
		}
		void PlayEndthemeSound()
		{
			Audio::Source src(*Sound::Buf::endtheme, Audio::Source::RepeatMode::single_keep);
			src.Play();
		}

		void Ending(WorldData &w)
		{
			if (w.ending)
				return;
			PlayEndthemeSound();
			w.ending = 1;
			w.ending_timer = 0;
		}

		void KillPlayer(WorldData &w)
		{
			if (w.player.dead == 0)
			{
				w.player.dead = 1;
				PlayDeathSound();
			}
		}

		void SpawnBullet(WorldData &w, M::ivec2 pos, M::ivec2 dir)
		{
			DynObjData tmp{5,0,pos,dir,1,0,0,{-1,-1},{}};
			w.dynamic_objs.push_back(tmp);
		}

		M::ivec2 TileCoordsForPos(M::ivec2 pos)
		{
			M::ivec2 ret = pos / 16;
			if (pos.x < 0) ret.x -= 1;
			if (pos.y < 0) ret.y -= 1;
			return ret;
		}

		unsigned int TileAtCoords(unsigned int world, M::ivec2 pos)
		{
			if (pos.x < 0) pos.x = 0;
			if (pos.y < 0) pos.y = 0;
			if (pos.x >= (int)Internal::worlds[world].w) pos.x = (int)Internal::worlds[world].w-1;
			if (pos.y >= (int)Internal::worlds[world].h) pos.y = (int)Internal::worlds[world].h-1;

			return Internal::worlds[world].map[pos.y * Internal::worlds[world].w + pos.x];
		}
		unsigned int TileAt(unsigned int world, M::ivec2 pos)
		{
			return TileAtCoords(world, TileCoordsForPos(pos));
		}

		unsigned int ObjAtCoords(unsigned int world, M::ivec2 pos)
		{
			if (pos.x < 0) pos.x = 0;
			if (pos.y < 0) pos.y = 0;
			if (pos.x >= (int)Internal::worlds[world].w) pos.x = (int)Internal::worlds[world].w-1;
			if (pos.y >= (int)Internal::worlds[world].h) pos.y = (int)Internal::worlds[world].h-1;
			return Internal::worlds[world].objmap[pos.y * Internal::worlds[world].w + pos.x];
		}
		unsigned int ObjAt(unsigned int world, M::ivec2 pos)
		{
			return ObjAtCoords(world, TileCoordsForPos(pos));
		}

		const Config::TileTypeData &TileInfoAt(unsigned int world, M::ivec2 pos)
		{
			return Config::list[TileAt(world, pos)];
		}
		const Config::ObjTypeData &ObjInfoAt(unsigned int world, M::ivec2 pos)
		{
			return Config::objlist[ObjAt(world, pos)];
		}

		bool SolidAt(unsigned int world, M::ivec2 pos)
		{
			if (pos.x < 0 || pos.y < 0 || pos.x >= (int)Internal::worlds[world].w * 16 || pos.y >= (int)Internal::worlds[world].h * 16)
				return 1;
			return TileInfoAt(world, pos).solid;
		}
		bool SolidAt_Player(unsigned int world, M::ivec2 pos)
		{
			for (M::ivec2 it : Config::player_hitbox)
			{
				if (SolidAt(world, pos + it))
					return 1;
			}
			return 0;
		}
		bool SolidAt_Obj(unsigned int world, M::ivec2 pos, const std::vector<M::ivec2> &hbox)
		{
			for (M::ivec2 it : hbox)
			{
				if (SolidAt(world, pos + it))
					return 1;
			}
			return 0;
		}

		void Save(WorldData &w)
		{
			if (w.player.dead)
				return;
			saved_world = w;
		}
		void Load(WorldData &w)
		{
			w = saved_world;
			w.campos = w.player.pos;
		}

		void Music(bool x)
		{
			if (x)
				Internal::theme_source->Play();
			else
				Internal::theme_source->Stop();
		}

		void DrawMap(unsigned int world, M::ivec2 pos)
		{
			M::ivec2 sz{WinW() / 16 + 2, WinH() / 16 + 2};
			M::ivec2 spos = TileCoordsForPos(pos - M::ivec2{WinWh(), WinHh()});
			M::ivec2 off = spos * 16 - pos;

			for (int y = 0; y < sz.y; y++)
			{
				for (int x = 0; x < sz.x; x++)
				{
					const Config::TileTypeData &cur = Config::list[TileAtCoords(world, {x + spos.x, y + spos.y})];

					if (!cur.visible)
						continue;
					cur.draw_func(M::ivec2{x, y} * 16 + off, world, {x + spos.x, y + spos.y});
				}
			}

			for (int y = 0; y < sz.y; y++)
			{
				for (int x = 0; x < sz.x; x++)
				{
					const Config::ObjTypeData &cur = Config::objlist[ObjAtCoords(world, {x + spos.x, y + spos.y})];
					if (!cur.visible)
						continue;
					cur.draw_func(M::ivec2{x, y} * 16 + off, world, {x + spos.x, y + spos.y});
				}
			}
		}

		void MovePlayer(WorldData &world, unsigned int worldnum, M::ivec2 pos, bool right = 1)
		{
			if (world.worldnum != worldnum)
			{
				world.dynamic_objs.clear();
			}
			world.worldnum = worldnum;
			world.player.pos = pos * 16 + M::ivec2{8, 16};
			world.campos = world.player.pos;
			world.player.facing_right = right;
		}

		void NextWorld(WorldData &world, M::ivec2 pos, M::fvec4 color, bool right = 1)
		{
			if (world.next_world_req)
				return;
			world.next_world_req = 1;
			world.next_world_timer = 0;
			world.next_world_loc = pos;
			world.next_world_fade_color = color;
			world.next_world_face_right = right;
		}

		void WorldTick(WorldData &w)
		{
			static constexpr float gravity = 0.091, unjump_gravity = 0.2, h_walk_s = 0.3, h_fly_s = 0.1, jump_v = 3, h_v_fade_s = 0.3, cam_move_v = 0.05;
			static constexpr M::ivec2 max_v {2,5}, cam_dir {48, -16};
			static constexpr float max_h_walk_v = 2;

			static constexpr int obj_margin = 128;

			auto GlobalTick = [&]()
			{
				if (w.fade_time > 0)
					w.fade_time -= w.fade_step;
				if (w.fade_time < 0)
					w.fade_time = 0;

				if (w.next_world_req)
				{
					if (!w.player.dead)
					{
						w.next_world_timer++;
						if (w.next_world_timer >= 60)
						{
							w.next_world_req = 0;
							w.next_world_timer = 60;
							Game::MovePlayer(w, w.worldnum + 1, w.next_world_loc, w.next_world_face_right);
							switch (w.worldnum)
							{
							  case 0:
								PlayTeleportSound();
								break;
							  case 1:
								PlayTeleportSound();
								break;
							  case 2:
								// Play nothing
								break;
							  case 3:
								// Play nothing
								break;
							}
						}
					}
					else
					{
						w.next_world_req = 0;
					}
				}
				else
				{
					if (w.next_world_timer > 0)
						w.next_world_timer--;
					if (w.next_world_timer < 0)
						w.next_world_timer = 0;
				}

				w.msg_text = "";


				if (w.worldnum == 2 && w.player.pos.x >= 127*16+8)
				{
					NextWorld(w, {1,36}, {1,1,1,1}, 1);
				}
				if (w.worldnum == 3 && w.player.pos.x >= int(Internal::worlds[3].w) * 16 - 8)
				{
					NextWorld(w, {1,42}, {1,1,1,1}, 1);
					Music(0);
					w.player.current_gun = 0;
				}
			};

			auto ObjectTick = [&]()
			{
				M::ivec2 sz{WinW() / 16 + 2, WinH() / 16 + 2};
				M::ivec2 spos = TileCoordsForPos(w.campos - M::ivec2{WinWh(), WinHh()});

				for (int y = 0; y < sz.y; y++)
				{
					for (int x = 0; x < sz.x; x++)
					{
						const Config::ObjTypeData &cur = Config::objlist[ObjAtCoords(w.worldnum, {x + spos.x, y + spos.y})];
						if (cur.spawn_dyn)
						{
							M::ivec2 mappos = spos + M::ivec2{x,y};
							bool ok = 1;
							for (const DynObjData &it : w.dynamic_objs)
							{
								if (it.home_loc == mappos)
								{
									ok = 0;
									break;
								}
							}
							if (ok)
							{
								DynObjData tmp{cur.spawn_dyn, 0, mappos * 16 + cur.spawn_off, {0,0}, cur.spawn_right, 0, 0, mappos, {}};
								w.dynamic_objs.push_front(tmp);
								if (!cur.respawn_if_dead)
								{
									DynObjData tmp{cur.spawn_once ? 6u : 0u, 0, mappos * 16 + cur.spawn_off, {0,0}, 0, 0, 0, mappos, {}};
									w.dynamic_objs.push_front(tmp);
								}
							}
						}
					}
				}

				for (auto it = w.dynamic_objs.begin(); it != w.dynamic_objs.end(); it++)
				{
					auto &p = *it;

					if (Config::dynobjlist[p.id].phys)
					{
						bool on_ground = SolidAt_Player(w.worldnum, {p.pos.x, p.pos.y + 1});

						// gravity
						if (!on_ground)
						{
							if (p.vel.y < 0)
								p.vel.y += unjump_gravity;
							else
								p.vel.y += gravity;
						}

						// velocity limit
						if (on_ground)
						{
							if (p.vel.x >  max_h_walk_v) p.vel.x =  max_h_walk_v;
							if (p.vel.x < -max_h_walk_v) p.vel.x = -max_h_walk_v;
						}
						else
						{
							if (p.vel.x >  max_v.x) p.vel.x =  max_v.x;
							if (p.vel.x < -max_v.x) p.vel.x = -max_v.x;
						}
						if (p.vel.y >  max_v.y) p.vel.y =  max_v.y;
						if (p.vel.y < -max_v.y) p.vel.y = -max_v.y;

						// pushing out of blocks
						if (SolidAt_Obj(w.worldnum, p.pos, Config::dynobjlist[p.id].hbox))
						{
							bool ok = 0;
							for (int i = 1; i < 16; i++)
							{
								if (!SolidAt_Obj(w.worldnum, {p.pos.x + i, p.pos.y}, Config::dynobjlist[p.id].hbox))
								{
									p.pos.x += i;
									ok = 1;
									break;
								}
								if (!SolidAt_Obj(w.worldnum, {p.pos.x - i, p.pos.y}, Config::dynobjlist[p.id].hbox))
								{
									p.pos.x -= i;
									ok = 1;
									break;
								}
								if (!SolidAt_Obj(w.worldnum, {p.pos.x, p.pos.y + i}, Config::dynobjlist[p.id].hbox))
								{
									p.pos.y += i;
									ok = 1;
									break;
								}
								if (!SolidAt_Obj(w.worldnum, {p.pos.x, p.pos.y - i}, Config::dynobjlist[p.id].hbox))
								{
									p.pos.y -= i;
									ok = 1;
									break;
								}
							}
							if (!ok)
								KillPlayer(w);
						}

						// pos modification
						if (p.vel.x)
						{
							int dir = p.vel.x > 0 ? 1 : -1;
							int v = std::round(p.vel.x);

							while (v)
							{
								if (!SolidAt_Player(w.worldnum, {p.pos.x+dir, p.pos.y}))
									p.pos.x += dir;
								else
									break;
								v -= dir;
							}
							if (SolidAt_Player(w.worldnum, {p.pos.x+dir, p.pos.y}))
								p.vel.x = 0;
						}
						if (p.vel.y)
						{
							int dir = p.vel.y > 0 ? 1 : -1;
							int v = std::round(p.vel.y);

							while (v)
							{
								if (!SolidAt_Player(w.worldnum, {p.pos.x, p.pos.y+dir}))
									p.pos.y += dir;
								else
									break;
								v -= dir;
							}
							if (SolidAt_Player(w.worldnum, {p.pos.x, p.pos.y+dir}))
								p.vel.y = 0;
						}
					}

					if (Config::dynobjlist[p.id].persistent == 0 && (p.pos.x > w.campos.x + WinWh() + obj_margin || p.pos.x < w.campos.x - WinWh() - obj_margin ||
																	 p.pos.y > w.campos.y + WinHh() + obj_margin || p.pos.y < w.campos.y - WinHh() - obj_margin))
					{
						it = w.dynamic_objs.erase(it);
					}

					if (p.dead)
					{
						if (p.id == 7)
						{
							w.aliens_killed++;
							if (w.aliens_killed >= 6)
							{
								StartDialogue(w, {"Our scanners show that you've just\nkilled the last alien on this base.",
												  "You did well, soldier.",
												  "Now go back to the teleporter and\nreturn back to the base."}, 1, {1,0.5,0.25,1});
							}
						}

						it = w.dynamic_objs.erase(it);
					}
				}
			};

			auto PlayerTick = [&]()
			{
				PlayerData &p = w.player;


				bool on_ground = SolidAt_Player(w.worldnum, {p.pos.x, p.pos.y + 1});
				bool can_control = (p.dead == 0 && w.dialogue == 0 && w.ending == 0);

				if (w.ending)
				{
					if (w.ending_timer < 160)
						w.ending_timer++;
					if (w.ending_timer > 160)
						w.ending_timer = 160;
				}

				if (can_control)
				{
					p.tex_off = {0,0};
					if (p.time_since_shot < 60)
						p.time_since_shot++;
					if (p.time_since_shot > 60)
						p.time_since_shot = 60;
				}

				// death timer
				if (p.dead)
				{
					p.death_timer++;
					if (p.death_timer > 300)
						p.death_timer = 300;
				}
				else
				{
					p.death_timer = 0;
				}

				// gravity
				if (!on_ground)
				{
					if (p.vel.y < 0 && Input::KeyUp(KeyJump()))
						p.vel.y += unjump_gravity;
					else
						p.vel.y += gravity;
				}

				// controls
				{
					// h controls
					int hc = can_control ? Input::KeyDown(KeyRight()) - Input::KeyDown(KeyLeft()) : 0;
					if (hc)
					{
						p.facing_right = hc > 0;
						if (std::abs(p.vel.x) < max_v.x)
							p.vel.x += hc * (on_ground ? h_walk_s : h_fly_s);
					}
					else
					{
						if (std::abs(p.vel.x) < h_v_fade_s)
							p.vel.x = 0;
						else
							p.vel.x += p.vel.x > 0 ? -h_v_fade_s : h_v_fade_s;
					}
					// v controls
					if (Input::KeyPressed(KeyJump()) && on_ground && can_control)
					{
						p.vel.y = -jump_v;
					}
					// gun controls
					p.aim_y = can_control ? Input::KeyDown(KeyDown()) - Input::KeyDown(KeyUp()) : 0;
					if (Input::KeyPressed(KeyShoot()) && can_control)
					{
						switch (p.current_gun)
						{
						  case 1:
						  	if (p.time_since_shot > 10)
						  	{
						  		p.time_since_shot = 0;
								PlayPistolSound();
								if (p.aim_y == -1)
									SpawnBullet(w, p.pos + (p.facing_right ? M::ivec2{-5,-18} : M::ivec2{4,-18}), {0,p.aim_y});
								else if (p.aim_y == 1)
									SpawnBullet(w, p.pos + (p.facing_right ? M::ivec2{3,10} : M::ivec2{-4,10}), {0,p.aim_y});
								else
									SpawnBullet(w, p.pos + (p.facing_right ? M::ivec2{13,-8} : M::ivec2{-14,-8}), {p.facing_right ? 1 : -1,0});
								break;
							}
						}
					}

					// respawn button
					if (p.dead && p.death_timer >= 30)
					{
						if (Input::KeyPressed(KeyShoot()) || Input::KeyPressed(KeyJump()))
						{
							Load(w);
							PlayLoadSound();
							Fade(w, 1, 0, {0,0.75f,1,1});
						}
					}
				}

				// velocity limit
				if (on_ground)
				{
					if (p.vel.x >  max_h_walk_v) p.vel.x =  max_h_walk_v;
					if (p.vel.x < -max_h_walk_v) p.vel.x = -max_h_walk_v;
				}
				else
				{
					if (p.vel.x >  max_v.x) p.vel.x =  max_v.x;
					if (p.vel.x < -max_v.x) p.vel.x = -max_v.x;
				}
				if (p.vel.y >  max_v.y) p.vel.y =  max_v.y;
				if (p.vel.y < -max_v.y) p.vel.y = -max_v.y;

				// pushing out of blocks
				if (SolidAt_Player(w.worldnum, p.pos))
				{
					bool ok = 0;
					for (int i = 1; i < 16; i++)
					{
						if (!SolidAt_Player(w.worldnum, {p.pos.x + i, p.pos.y}))
						{
							p.pos.x += i;
							ok = 1;
							break;
						}
						if (!SolidAt_Player(w.worldnum, {p.pos.x - i, p.pos.y}))
						{
							p.pos.x -= i;
							ok = 1;
							break;
						}
						if (!SolidAt_Player(w.worldnum, {p.pos.x, p.pos.y + i}))
						{
							p.pos.y += i;
							ok = 1;
							break;
						}
						if (!SolidAt_Player(w.worldnum, {p.pos.x, p.pos.y - i}))
						{
							p.pos.y -= i;
							ok = 1;
							break;
						}
					}
					if (!ok)
						KillPlayer(w);
				}

				// pos modification
				if (p.vel.x)
				{
					int dir = p.vel.x > 0 ? 1 : -1;
					int v = std::round(p.vel.x);

					while (v)
					{
						if (!SolidAt_Player(w.worldnum, {p.pos.x+dir, p.pos.y}))
							p.pos.x += dir;
						else
							break;
						v -= dir;
					}
					if (SolidAt_Player(w.worldnum, {p.pos.x+dir, p.pos.y}))
						p.vel.x = 0;
				}
				if (p.vel.y)
				{
					int dir = p.vel.y > 0 ? 1 : -1;
					int v = std::round(p.vel.y);

					while (v)
					{
						if (!SolidAt_Player(w.worldnum, {p.pos.x, p.pos.y+dir}))
							p.pos.y += dir;
						else
							break;
						v -= dir;
					}
					if (SolidAt_Player(w.worldnum, {p.pos.x, p.pos.y+dir}))
						p.vel.y = 0;
				}

				// object interactrion
				{
					bool kill_p = 0;
					M::ivec2 arr[] {{0,-8}, {0,8}, {-1,-8}, {-1,8}};
					for (M::ivec2 it : arr)
					{
						unsigned int id = ObjAt(w.worldnum, p.pos+it);
						switch (id)
						{
						  case 4:
						  case 8:
							kill_p = 1;
							break;
						  default:
							break;
						}
					}
					if (kill_p)
						KillPlayer(w);
				}

				// dyn object interaction
				if (p.dead == 0 && w.dialogue == 0)
				{
					bool play_field_sound = 0;

					for (auto it = w.dynamic_objs.begin(); it != w.dynamic_objs.end(); it++)
					{
						switch (it->id)
						{
						  case 0:
						  default:
							break;
						  case 1: // check point
						  	{
								if (p.pos.y == it->pos.y && p.pos.x >= it->pos.x-10 && p.pos.x < it->pos.x+10)
									p.tex_off.y = -2;
								if (p.pos.y == it->pos.y && p.pos.x >= it->pos.x-8 && p.pos.x < it->pos.x+8)
								{
									DrawMessage(w, Jo("Press ", KeyName_Down(), " to save"));
									if (Input::KeyPressed(KeyDown()))
									{
										Fade(w, 1, 0, {1,1,1,1});
										PlaySaveSound();
										Save(w);
									}
									if (Input::KeyDown(KeyDown()))
										p.aim_y = 0;
								}
							}
							break;
						  case 2: // base girl
						  	{
								it->facing_right = p.pos.x > it->pos.x;
								if (it->data.size() == 0 && p.pos.x >= it->pos.x - 64)
								{
									p.facing_right = 1;
									StartDialogue(w, {"Welcome, soldier!",
													  "I have an important mission for you\ntoday.",
													  "",
													  "Yes?",
													  "",
													  "We've found an another base of these\ndamned aliens!",
													  "Your mission is to go there, investigate\nthat base and kill every living thing in\nit.",
													  "Be careful, they are dangerous!",
													  "Use the teleporter to the right of me.",
													  "Good luck!"}, 0, {1,0.5,0.25,1});
									it->data.push_back({0,0});
								}
							}
							break;
						  case 3: // teleporter
						  	{
								if (p.pos.y == it->pos.y && p.pos.x >= it->pos.x-13 && p.pos.x < it->pos.x+13)
									p.tex_off.y = -3;
								if (p.pos.y == it->pos.y && p.pos.x >= it->pos.x-10 && p.pos.x < it->pos.x+10)
								{
									DrawMessage(w, Jo("Press ", KeyName_Down(), " to use"));
									if (Input::KeyPressed(KeyDown()))
									{
										p.current_gun = 1;
										NextWorld(w, {12,40}, {1,1,0.9,1}, 1);
									}
									if (Input::KeyDown(KeyDown()))
										p.aim_y = 0;
								}
							}
							break;
						  case 4: // fake teleporter
						  	{
								if (p.pos.y == it->pos.y && p.pos.x >= it->pos.x-13 && p.pos.x < it->pos.x+13)
									p.tex_off.y = -3;
								if (p.pos.y == it->pos.y && p.pos.x >= it->pos.x-10 && p.pos.x < it->pos.x+10)
								{
									DrawMessage(w, Jo("Press ", KeyName_Down(), " to use"));
									if (Input::KeyPressed(KeyDown()))
									{
										if (w.aliens_killed < 6)
											StartDialogue(w, {"Our scanners show that they are not dead\nyet.",
															  "Go back and finish your job!"}, 1, {1,0.5,0.25,1});
										else
											StartDialogue(w, {"", "It does not work."}, 1, {1,0.5,0.25,1});
									}
									if (Input::KeyDown(KeyDown()))
										p.aim_y = 0;
								}
							}
							break;
						  case 5: // - player bullet
							{
								it->pos += it->vel * 4;
								if (SolidAt(w.worldnum, it->pos) || ObjAt(w.worldnum, it->pos) == 8)
									it = w.dynamic_objs.erase(it);
							}
							break;
						  case 6: // - pers. marker
							break;
						  case 7: // alien
							{
								if (it->data.size() == 0 && p.pos.y >= it->pos.y - 48 && p.pos.y < it->pos.y + 48 && std::abs(p.pos.x - it->pos.x) < 175)
								{
									it->data.push_back({0,0});
								}
								if (it->data.size() != 0)
								{
									it->facing_right = p.pos.x < it->pos.x;
									it->vel.x = it->facing_right ? 1 : -1;
								}

								for (auto jt = w.dynamic_objs.begin(); jt != w.dynamic_objs.end(); jt++)
								{
									if (jt->id != 5)
										continue;
									if (jt->pos.x >= it->pos.x-6 && jt->pos.x < it->pos.x+6 && jt->pos.y >= it->pos.y-12 && jt->pos.y < it->pos.y+16)
									{
										PlayMobdeathSound();
										it->dead = 1;
										jt->dead = 1;
										break;
									}
								}
							}
							break;
						  case 8: // oldman
							{
								it->facing_right = p.pos.x > it->pos.x;
								if (it->data.size() == 0 && p.pos.x <= it->pos.x + 40)
								{
									p.facing_right = it->pos.x > p.pos.x;
									StartDialogue(w, {"",
													  "Who are you?",
													  "What are doing here?",
													  "",
													  "...",
													  "Once I was a soldier like you...\n",
													  "",
													  "You? Soldier? Funny stories you have\nthere. You are too decrepit to even hold\na gun!",
													  "",
													  "Be quiet and listen! My words can save\nyour life!",
													  "So, I was the previous one. The hero,\nslayer of aliens...",
													  "But one day I became too old, too smart...\nAnd they 'fired' me and left me in this\nplace, just as they did with you.",
													  "I know the way to escape, but I'm too\nold, too weak to use it.",
													  "I will tell you. It's you who shall\nescape, not me.",
													  "But before that, I must give you one\nsingle advice: Don't blame yourself\nafter, you have known nothing...",
													  "Now I will share my knowledge with you.",
													  "I will show you...",
													  "The Truth!"}, 2, {0,0.5,0.5,1});
									it->data.push_back({0,0});
								}
								if (w.dialogue == 0 && it->data.size() == 1)
								{
									NextWorld(w, {12,40}, {0.85,1,0.85,1}, 0);
									it->data.push_back({0,0});
								}
							}
							break;
						  case 9: // turret
							{
								if (it->data.size() == 0)
								{
									it->facing_right = p.pos.x > it->pos.x;
									it->data.push_back({0,0});
								}
								if (it->facing_right != (p.pos.x > it->pos.x))
								{
									it->facing_right = p.pos.x > it->pos.x;
									PlayRotSound();
								}
								if (p.pos.y >= it->pos.y-8 && p.pos.y < it->pos.y+8 && p.pos.x >= it->pos.x-8 && p.pos.x < it->pos.x+8)
								{
									KillPlayer(w);
								}

								if (p.pos.y >= it->pos.y-32 && p.pos.y < it->pos.y+32)
								{
									if (it->data.size() < 2)
									{
										it->data.push_back({it->facing_right ? it->pos.x + 10 : it->pos.x - 10, it->pos.y-7});
										PlayTshotSound();
									}
								}
								if (it->data.size() == 2)
								{
									M::ivec2 &ref = *(it->data.end()-1);
									ref.x += (it->facing_right ? 4 : -4);
									if (ref.x >= p.pos.x-8 && ref.x < p.pos.x+8 && ref.y >= p.pos.y-15 && ref.y < p.pos.y+16)
										KillPlayer(w);
									if (SolidAt(w.worldnum, ref))
										it->data.erase(it->data.end()-1);
								}

								for (auto jt = w.dynamic_objs.begin(); jt != w.dynamic_objs.end(); jt++)
								{
									if (jt->id != 5)
										continue;
									if (jt->pos.x >= it->pos.x-6 && jt->pos.x < it->pos.x+6 && jt->pos.y >= it->pos.y-12 && jt->pos.y < it->pos.y+16)
									{
										PlayMobdeathSound();
										it->dead = 1;
										jt->dead = 1;
										break;
									}
								}
							}
							break;
						  case 10: // forcefield
						  	{
						  		if (w.time % 60 == 0 && !play_field_sound)
									play_field_sound = 1;

						  		if (p.pos.x >= it->pos.x-9 && p.pos.x < it->pos.x+9 && p.pos.y >= it->pos.y-9 && p.pos.y < it->pos.y+9 && w.time % 120 < 60)
									KillPlayer(w);
						  	}
							break;
						}
					}

					if (play_field_sound)
						PlayFieldSound();
				}

				// camera movement
				M::ivec2 targetcampos;
				if (p.aim_y == 0)
				{
					targetcampos = {p.pos.x + int(p.facing_right ? cam_dir.x : -cam_dir.x), p.pos.y + cam_dir.y};
				}
				else
				{
					targetcampos = {p.pos.x, p.pos.y + cam_dir.y + int(p.aim_y > 0 ? cam_dir.x : -cam_dir.x)};
				}
				M::fvec2 cam_off = targetcampos - w.campos;
				w.campos += M::ivec2{int(std::round(cam_off.x * cam_move_v)), int(std::round(cam_off.y * cam_move_v))};

				// anim frame selection
				if (p.vel.x && on_ground)
				{
					p.anim_frame = p.walk_time / 5 % 10 + 1;
					p.walk_time = (p.walk_time + 1) % 50;
				}
				else
				{
					p.walk_time = 0;
					if (on_ground)
					{
						p.anim_frame = 0;
					}
					else
					{
						if (p.vel.y < 0)
							p.anim_frame = 11;
						else
							p.anim_frame = 12;
					}
				}
			};

			auto GuiTick = [&]()
			{
				if (w.player.pos.x >= 109*16+8 && w.worldnum == 1 && w.talked_with_center == 0)
				{
					w.talked_with_center = 1;
					StartDialogue(w, {"",
									  "Center, they seem pretty peaceful, maybe\nwe could just talk with them?",
									  "",
									  "Shut up and do your job!"},1,{1,0.5,0.25,1});
				}

				if (w.player.pos.x < 44*16+8 && w.worldnum == 1 && w.talked_with_center_2 == 0 && w.aliens_killed >= 6)
				{
					DynObjData tmp{8, 0, {9*16+8,40*16},{0,0},1,0,0,{-1,-1},{}};
					w.dynamic_objs.push_back(tmp);
					w.talked_with_center_2 = 1;
					StartDialogue(w, {"Thanks for you job again, soldier!",
									  "I'm sorry, but our plans have changed.",
									  "",
									  "What?",
									  "",
									  "We don't need you anymore.",
									  "We've turned the teleporter off.",
									  "",
									  "What the?..",
									  "",
									  "May your bones rest in peace on this\nforgotten planet.",
									  "Over and out."}, 1, {1,0.5,0.25,1});
				}

				if (w.worldnum == 2)
				{
					if (w.wowmsg_0 == 0 && w.player.pos.x >= 15*16+8)
					{
						w.wowmsg_0 = 1;
						StartDialogue(w, {"",
										  "What the?!",
										  "Where am I?",}, 0, {0,0,0,0});
					}
					if (w.wowmsg_1 == 0 && w.player.pos.x >= 37*16)
					{
						w.wowmsg_1 = 1;
						StartDialogue(w, {"",
										  "This place looks surpisingly similar to\nthe planet...",
										  "And it looks like the base interior...",
										  "...",
										  "That planet was an illulsion the whole\ntime?!",
										  "...",
										  "I wonder who were those aliens then..."}, 0, {0,0,0,0});
					}
					if (w.wowmsg_2 == 0 && w.player.pos.x >= 84*16)
					{
						w.wowmsg_2 = 1;
						StartDialogue(w, {"",
										  "All those dead bodies...",
										  "Was it me who killed them?",
										  "...",
										  "I have very serious questions to the base\ndirector..."}, 0, {0,0,0,0});
					}
				}
				if (w.worldnum == 4)
				{
					if (w.girlmsg_0 == 0 && w.player.pos.x >= 14*16+8)
					{
						w.girlmsg_0 = 1;
						StartDialogue(w, {"What the?..",
										  "How?.."}, 0, {1,0.5,0.25,1});
					}
					if (w.girlmsg_1 == 0 && w.player.pos.x >= 18*16+8)
					{
						w.girlmsg_1 = 1;
						StartDialogue(w, {"I can explain it!",
										  "They were militay prisoners, it was\nnecessary to...",
										  "",
										  "Silence!",
										  "I don't want to hear this shit."}, 0, {1,0.5,0.25,1});
					}
					if (w.girlmsg_2 == 0 && w.player.pos.x >= 26*16+8)
					{
						w.girlmsg_2 = 1;                     // I have very serious questions to the base
						StartDialogue(w, {"I thought that you will kill me...",
										  "...",
										  "",
										  "...",
										  "",
										  "Thank you."}, 0, {1,0.5,0.25,1});
					}
				}

				if (w.hint_txt)
				{
					w.hint_timer--;
					if (w.hint_timer <= 0)
					{
						w.hint_txt = 0;
						w.hint_timer = 0;
					}
					else
					{
						if (w.hint_timer > 60)
							w.hint_timer = 60;
					}
				}

				if (w.dialogue)
				{
					if (!*w.dia_rem_text)
					{
						if (Input::KeyPressed(Game::KeyJump()) || Input::KeyPressed(KeyShoot()))
						{
							w.dia_timer = 30;
							w.dia_text = "";

							do
							{
								w.dia_rem_pages.erase(w.dia_rem_pages.begin());
								if (w.dia_rem_pages.size() == 0)
								{
									w.dialogue = 0;
									w.dia_text = "";
									break;
								}
								else
									w.dia_rem_text = w.dia_rem_pages.begin()->c_str();

								if (*w.dia_rem_text == 0)
									w.dia_self_speech = !w.dia_self_speech;
							}
							while (*w.dia_rem_text == 0);
						}
					}
					else
					{
						if (w.dia_timer <= 0)
						{
							char tmp = *(w.dia_rem_text++);
							w.dia_text += tmp;
							w.dia_timer = 3;
							if (tmp != ' ' && tmp != '\n' && tmp != ',' && tmp != '.')
								Game::PlayDiaSound();
						}
						else
						{
							if (Input::KeyDown(KeyJump()) || Input::KeyDown(KeyShoot()))
								w.dia_timer -= 10;
							else
								w.dia_timer--;
						}
					}
				}
			};

			GlobalTick();
			ObjectTick();
			PlayerTick();
			GuiTick();

			w.time++;
		}

		void WorldRender(WorldData &w)
		{
			auto GlobalRender = [&]()
			{
				if (Internal::world_backgrounds[w.worldnum] != -1)
					DrawBackground(Internal::world_backgrounds[w.worldnum]);
				DrawMap(w.worldnum, w.campos);
			};

			auto ObjectRender = [&]()
			{
				for (auto &it : w.dynamic_objs)
				{
					if (it.id == 10 && w.time % 120 >= 60) // forcefiled
						continue;
					if (Config::dynobjlist[it.id].visible)
						Config::dynobjlist[it.id].draw_func(it.pos - w.campos, it);
					if (it.id == 9 && it.data.size() == 2) // turret
					{
						Draw::Rect(*(it.data.end()-1) - M::ivec2{8,8} - w.campos, {16,16}, {192,64});
					}
				}
			};

			auto PlayerRender = [&]()
			{
				PlayerData &p = w.player;
				M::ivec2 ppos = p.pos - w.campos + p.tex_off;
				M::ivec2 texpos {96 + p.anim_frame*24,80};

				if (p.dead)
					Draw::Color({1,1,1,1.0f - p.death_timer / 60.0f});

				Draw::Rect(ppos - M::ivec2{p.facing_right ? 12 : -12,16},{p.facing_right ? 24 : -24,32},texpos, {24, 32});

				float a = (p.current_gun != 0 ? p.aim_y : 0) * (p.facing_right ? 1 : -1) * fPi_half;
				Draw::RectRot(ppos - M::ivec2{p.facing_right ? 1 : 0,4} + M::fvec2{0.5,0.5},{p.facing_right ? 24 : -24,24},{96 + p.current_gun*24, 112}, {24, 24}, a, {7.5,12.5});

				if (p.dead)
					Draw::Color();

				if (p.dead)
				{
					float t = p.death_timer / 300.0f * 0.85f;
					Draw::GlobalColor({1.0f - t,1.0f - t * 2,1.0f - t * 4,1});
				}
			};

			auto GuiRender = [&]()
			{
				if (w.hint_timer)
				{
					int ww = (WinW() + 31) / 32;
					Draw::Color({1,1,1,w.hint_timer / 60.0f});
					for (int i = 0; i < ww; i++)
						Draw::Rect({-WinWh() + i*32, -WinHh() + 32}, {32,32}, {128,0});
					Draw::Color();
					M::fvec4 tmpc = w.hint_color;
					tmpc.a *= w.hint_timer / 60.0f;
					Draw::Text({0, -WinHh() + 32 + 10}, w.hint_txt, tmpc, 0, {0,1}, {0,0,0,w.hint_timer / 60.0f});
				}

				if (w.dialogue)
				{
					Draw::Rect({-WinWh(), -WinHh()}, {64,64}, {192,0});
					int ww = (WinW() - 64 + 31) / 32;
					for (int i = 0; i < ww; i++)
						Draw::Rect({-WinWh()+64+i*32, -WinHh()}, {32,64}, {256, 0});
					if (!w.dia_self_speech)
						Draw::Rect({-WinWh()+8, -WinHh()+8}, {48,48}, {288 + int(w.dia_icon) * 48,0});
					M::fvec4 color = w.dia_self_speech ? M::fvec4{0.75,0.75,0.75,1} : w.dia_color;
					Draw::Text({-WinWh()+70-3, -WinHh()+8}, w.dia_text.c_str(), color, -1, {1,1});
					if (!*w.dia_rem_text)
						Draw::Text({WinWh()-6, -WinHh()+64-8-12}, Jo(KeyName_Jump(), " / ", KeyName_Shoot()), {1,1,1,1}, 1, {1,1});
				}

				if (w.msg_text.c_str()[0] != '\0')
				{
					for (int y : {-1,1,0})
					{
						for (int x : {-1,1,0})
						{
							Draw::Text({x, y+48}, w.msg_text.c_str(), x == 0 && y == 0 ? M::fvec4{0.75,0.75,0.75,1} : M::fvec4{0,0,0,1}, 0);
						}
					}
				}
			};

			auto PostRender = [&]()
			{
				if (w.fade_time)
				{
					Draw::Color({w.fade_color.x,w.fade_color.y,w.fade_color.z,w.fade_time * w.fade_color.a});
					Draw::RectFill({-WinWh(), -WinHh()}, {WinW(), WinH()}, {1,0});
					Draw::Color();
				}

				if (w.next_world_timer)
				{
					M::fvec4 c = w.next_world_fade_color;
					c.w *= w.next_world_timer / 50.0f;
					Draw::Color(c);
					Draw::RectFill({-WinWh(), -WinHh()}, {WinW(), WinH()}, {1,0});
					Draw::Color();
				}

				if (w.player.dead)
				{
					float a = -0.25f + w.player.death_timer / 300.0f * 1.25f;
					Draw::Color({1,1,1,a});
					Draw::Rect({-104,-16},{208,32},{0,440});
					Draw::Color();
					Draw::Text({0,40}, Jo("Press ", KeyName_Jump(), " or ", KeyName_Shoot(), " to start from the last checkpoint"), {0.75,0,0,a}, 0);
				}

				if (w.ending)
				{
					Draw::Color({1,1,1,w.ending_timer / 120.0f});
					Draw::Rect({-192/2,-110}, {192,48}, {96,144});
					Draw::Color();
				}
			};

			GlobalRender();
			ObjectRender();
			PlayerRender();
			GuiRender();

			ForceRender();
			Draw::GlobalColor();

			PostRender();
		}
	}



	void PreTick()
	{
		System::PreTick();
		if (Input::MouseFocus())
		{
			Internal::mx = (Input::MouseX() + Window::W_half() * (Internal::scale - 1)) / Internal::scale - Window::W_half();
			Internal::my = (Input::MouseY() + Window::H_half() * (Internal::scale - 1)) / Internal::scale - Window::H_half();
		}
		else
		{
			Internal::mx = 9000;
			Internal::my = 0;
		}
		ren_queue->Clear();
	}

	void PostTick()
	{
		ren_queue->Send();
		ren_queue->Draw();
		System::PostTick();
	}

	namespace States
	{
		Main::StateFunc Boot()
		{
			// GFX
			master_shader.Ctor<const char *, const char *, const std::initializer_list<const char *> &>(ShaderCode::Master::v, ShaderCode::Master::f, {"in_p", "in_t", "in_cm", 0, 0, 0});
			master_shader->Use();
			UniformLocs::u_mat      = glGetUniformLocation(*master_shader, "u_mat"     );
			UniformLocs::u_texsz    = glGetUniformLocation(*master_shader, "u_texsz"   );
			UniformLocs::u_scale    = glGetUniformLocation(*master_shader, "u_scale"   );
			UniformLocs::u_tex      = glGetUniformLocation(*master_shader, "u_tex"     );
			UniformLocs::u_mastercm = glGetUniformLocation(*master_shader, "u_mastercm");

			CheckScale();
			Draw::GlobalColor();

			ren_queue.Ctor(1024*64*6);

			cam_control.ViewParams_Flat(-1, 1);
			cam_control.ApplyRatioToMatrixes(Window::W_half()*2, Window::H_half()*2);
			cam_control.SetUniView(UniformLocs::u_mat);

			GFX::ImageData img(C_ASSET("gfx",".tga"));
			glUniform2f(UniformLocs::u_texsz, img.W(), img.H());
			master_tex.Ctor();
			master_tex->SetActiveSlot(0);
			master_tex->BindToActiveSlot();
			master_tex->ActiveSlotData(img);
			master_tex->ActiveSlotParams(GFX::Texture::Interpolation::nearest, GFX::Texture::Wrap::transp);
			master_tex->SetSlotAsUniform(UniformLocs::u_tex, 0);

			// Audio
			Audio::ListenerPos({0, 0, -240});
			Audio::ListenerRot({0,0,1},{0,-1,0});

			// Loading controls
			Internal::keys[0] = Input::Key_Left();
			Internal::keys[1] = Input::Key_Right();
			Internal::keys[2] = Input::Key_Up();
			Internal::keys[3] = Input::Key_Down();
			Internal::keys[4] = Input::Key<'Z'>();
			Internal::keys[5] = Input::Key<'X'>();

			std::ifstream inp_f("controls_config.txt");
			if (inp_f)
			{
				for (unsigned int &it : Internal::keys)
					inp_f >> it;
			}
			inp_f.close();
			Game::SaveKeyConfig();
			Game::UpdateKeyNames();

			// Loading sounds
			Sound::Buf::err     .Ctor(ASSET("err.wav"     ), 0);
			Sound::Buf::ok      .Ctor(ASSET("ok.wav"      ), 0);
			Sound::Buf::dia     .Ctor(ASSET("dia.wav"     ), 0);
			Sound::Buf::death   .Ctor(ASSET("death.wav"   ), 1);
			Sound::Buf::save    .Ctor(ASSET("save.wav"    ), 0);
			Sound::Buf::load    .Ctor(ASSET("load.wav"    ), 0);
			Sound::Buf::teleport.Ctor(ASSET("teleport.wav"), 0);
			Sound::Buf::pistol  .Ctor(ASSET("pistol.wav"  ), 0);
			Sound::Buf::mobdeath.Ctor(ASSET("mobdeath.wav"), 0);
			Sound::Buf::rot     .Ctor(ASSET("rot.wav"     ), 0);
			Sound::Buf::tshot   .Ctor(ASSET("tshot.wav"   ), 0);
			Sound::Buf::field   .Ctor(ASSET("field.wav"   ), 0);
			Sound::Buf::endtheme.Ctor(ASSET("endtheme.wav"), 1);

			Sound::Buf::maintheme.Ctor(ASSET("maintheme.wav" ), 1);

			Internal::theme_source.Ctor<const Audio::Buffer &>(*Sound::Buf::maintheme, Audio::Source::RepeatMode::loop, 0.5f);

			// Loading levels
			for (int i = 0; i < Internal::worlds_c; i++)
			{
				std::string fname = Jo(ASSETS_DIR, "world", i);
				GFX::ImageData wrld(Utils::CheckIfFileExists(fname.c_str()) ? fname.c_str() : (fname + ".tga").c_str());
				fname = Jo(ASSETS_DIR, "data", i);
				GFX::ImageData wrld_o(Utils::CheckIfFileExists(fname.c_str()) ? fname.c_str() : (fname + ".tga").c_str());

				if (wrld.W() != wrld_o.W() || wrld.H() != wrld_o.H())
					Main::Exit(Jo("Size mismatch for tile and object maps for world ", i, '!'));

				Internal::worlds[i].w = wrld.W();
				Internal::worlds[i].h = wrld.H();
				Internal::worlds[i].map = new unsigned int[wrld.W() * wrld.H()];
				Internal::worlds[i].objmap = new unsigned int[wrld.W() * wrld.H()];

				for (unsigned int y = 0; y < wrld.H(); y++)
				{
					for (unsigned int x = 0; x < wrld.W(); x++)
					{
						const unsigned char *ptr = wrld.Data()[y * wrld.W() + x];
						M::ucvec4 src = {ptr[0], ptr[1], ptr[2], ptr[3]};
						for (const auto &it : Config::list)
						{
							if (it.color_code == src)
							{
                                Internal::worlds[i].map[y * wrld.W() + x] = &it - Config::list;
								goto stop_search;
							}
						}
						Main::Exit(Jo("Can't load world ",i,": Invalid tile at ",x,':',y," with code {",int(src.x),',',int(src.y),',',int(src.z),',',int(src.w),"}!"));
					  stop_search:

					  	ptr = wrld_o.Data()[y * wrld.W() + x];
					  	src = {ptr[0], ptr[1], ptr[2], ptr[3]};
					  	for (const auto &it : Config::objlist)
						{
							if (it.color_code == src)
							{
                                Internal::worlds[i].objmap[y * wrld.W() + x] = &it - Config::objlist;
								goto stop_search_0;
							}
						}
						Main::Exit(Jo("Can't load world ",i,": Invalid object at ",x,':',y," with code {",int(src.x),',',int(src.y),',',int(src.z),',',int(src.w),"}!"));
					  stop_search_0:
						;
					}
				}
			}

			return Menu;
		}

		Main::StateFunc Menu()
		{
			Game::Music(0);

			Input::MouseMode::Show();

			int selection = -1;

			Main::StateFunc (*retptr)() = 0;

			auto Tick = [&]
			{
				if (MouseX() >= -48 && MouseX() < 48 && MouseY() >= 16 && (MouseY() - 16) % 32 < 24 && (MouseY() - 16) / 32 < 3)
				{
					selection = (MouseY() - 16) / 32;
					if (Input::MouseButtonPressed(0))
					{
						switch (selection)
						{
						  case 0:
							retptr = Game;
							//Game::PlayOkSound();
							break;
						  case 1:
							retptr = Controls;
							Game::PlayOkSound();
							break;
						  case 2:
						  default:
							Main::Exit();
							break;
						}
					}
				}
				else
				{
					selection = -1;
				}
			};

			float hlight[3] {};

			auto Render = [&]
			{
				Game::DrawFlatBack();

				int pos = 0;
				for (const char *it : {"Start", "Controls", "Quit"})
				{
					static constexpr float step = 0.1;
					if (selection == pos)
					{
						hlight[pos] += step;
						if (hlight[pos] > 1)
							hlight[pos] = 1;
					}
					else
					{
						hlight[pos] -= step;
						if (hlight[pos] < 0)
							hlight[pos] = 0;
					}

					Draw::Color({1.0f - hlight[pos],1.0f - hlight[pos]/4.0f,1,1});
					Draw::Rect({-48,pos*32+16}, {96,24}, {96,32});
					Draw::RectFill({-48,pos*32+24+16}, {96,2}, {96,32});
					Draw::Color();
					Draw::Text({0, pos*32+16+6}, it, {0,0,0,1}, 0, {0,1}, {0,0,0,0.25});
					pos++;
				}

				Draw::Rect({-160, -120}, {320, 128}, {0, 192});
				Draw::Rect({-WinWh(), -120}, {WinWh() - 160, 128}, {1, 192}, {1, 128});
				Draw::Rect({160, -120}, {WinWh() - 160, 128}, {1, 192}, {1, 128});
			};

			while (1)
			{
				PreTick();

				Tick();
				Render();
				Draw::Text({0, WinHh()-12}, "A game by HolyBlackCat, made for LD48 #33 in 48 hours", {0,0,0,1}, 0);

				PostTick();

				if (retptr)
					return retptr;
			}
		}

		Main::StateFunc Game()
		{
			Game::Music(1);

			Input::MouseMode::Hide();

			WorldData world{};

			Game::MovePlayer(world, 0, {5, 22});

			Game::Save(world);

			Game::Fade(world, 4, 1, {0,0,0,1});

			auto PosHint = [&](M::fvec4 color, unsigned int w, int a, int b, const char *txt)
			{
				if (world.worldnum == w && world.player.pos.x >= a * 16 && world.player.pos.x < b * 16 + 16)
					Game::DrawHint(world, txt, color);
			};

			while (1)
			{
				PreTick();
				Game::WorldTick(world);

				Game::WorldRender(world);

				PosHint({1,0.5,0,1}, 0, 0, 8, Jo("Use ", Game::KeyName_Left(), " and ", Game::KeyName_Right(), " to move"));
				PosHint({1,0.5,0,1}, 0, 13, 14, Jo("Use ", Game::KeyName_Jump(), " to jump"));
				PosHint({1,0.5,0,1}, 0, 19, 20, Jo("Hold ", Game::KeyName_Jump(), " to jump higher"));
				PosHint({1,0.5,0,1}, 0, 36, 44, Jo("Use ", Game::KeyName_Up(), " and ", Game::KeyName_Down(), " to aim and look around"));
				PosHint({1,0.5,0,1}, 0, 50, 56, Jo("It is a checkpoint, use it to save your progress"));
				if (world.talked_with_center_2 == 0)
					PosHint({1,0.5,0,1}, 1,  5, 21, Jo("Press ", Game::KeyName_Shoot(), " to shoot"));

				if (world.worldnum == 4 && world.player.pos.x >= 16*60)
					Game::Ending(world);

				PostTick();
			}

			return Menu;
		}

		Main::StateFunc Controls()
		{
			Game::Music(0);

			Input::MouseMode::Show();

			int selection = -1, cursor = -1;

			Main::StateFunc (*retptr)() = 0;

			unsigned int timer = -1;

			auto Tick = [&]
			{
				if (MouseY() >= -108 && (MouseY() + 108) % 32 < 24 && (MouseY() + 108) / 32 < 7)
				{
					selection = (MouseY() + 108) / 32;
					if (selection == 6)
					{
						if (!(MouseX() >= -48 && MouseX() < 48))
						{
							selection = -1;
						}
					}
					else
					{
						if (!(MouseX() >= -96 && MouseX() < 0))
						{
							selection = -1;
						}
					}

					if (selection != -1 && Input::MouseButtonPressed(0))
					{
						if (selection == 6)
						{
							retptr = Menu;
							Game::PlayOkSound();
							Game::SaveKeyConfig();
							Game::UpdateKeyNames();
						}
						else
						{
							if (cursor == -1)
							{
								cursor = selection;
								timer = -1;
							}
						}
					}
				}
				else
				{
					selection = -1;
				}

				if (cursor != -1)
				{
					timer++;
					unsigned int tmp = 0;
					if (Input::WhatKeyIsPressed(&tmp))
					{
						bool ok = 1;
						for (int i = 0; i < 6; i++)
						{
							if (cursor == i)
								continue;
							if (Internal::keys[i] == tmp)
								ok = 0;
						}
						if (ok)
						{
							Internal::keys[cursor] = tmp;
							Game::UpdateKeyName(cursor);
							cursor = -1;
							Game::PlayOkSound();
						}
						else
						{
							Game::PlayErrorSound();
						}
					}
				}
			};

			float hlight[7] {};

			auto Render = [&]
			{
				Game::DrawFlatBack();

				int pos = 0;
				for (const char *it : {"Left", "Right", "Up", "Down", "Jump", "Shoot", "Back"})
				{
					static constexpr float step = 0.1;
					if (selection == pos)
					{
						hlight[pos] += step;
						if (hlight[pos] > 1)
							hlight[pos] = 1;
					}
					else
					{
						hlight[pos] -= step;
						if (hlight[pos] < 0)
							hlight[pos] = 0;
					}

					Draw::Color({1.0f - hlight[pos],1.0f - hlight[pos]/4.0f,1,1});
					Draw::Rect({-96 + (pos / 6 * 48),-108+pos*32}, {96,24}, {96,32});
					Draw::RectFill({-96 + (pos / 6 * 48),-108+pos*32+24}, {96,2}, {96,32});
					Draw::Color();
					Draw::Text({-48 + (pos / 6 * 48), -108+pos*32+6}, it, {0,0,0,1}, 0, {0,1}, {0,0,0,0.25});
					if (pos != 6)
					{
						Draw::Rect({0,-108+pos*32}, {96,24}, {96,56});
						Draw::RectFill({0,-108+pos*32+24}, {96,2}, {96,56});
						Draw::Text({48,-108+pos*32+6}, Internal::keynames[pos].c_str(), cursor == pos ? M::fvec4{0,timer % 60 >= 30 ? 0.75f : 0.5f,0,1} : M::fvec4{0.75f,0.75f,0.75f,1}, 0, {1,1},{0,0,0,1});
					}
					pos++;
				}
			};

			while (1)
			{
				PreTick();

				Tick();
				Render();

				PostTick();

				if (retptr)
					return retptr;
			}
		}
	}
}
