#include <cmath>

#include "system.h"

#include <ctime>
#include <cstring>
#include <iostream>
#include <map>
#include <string>

#define AUDIO_H_MAINT_ACCESS
#include "audio.h"
#include "system.h"
#define GFX_H_MAINT_ACCESS
#include "gfx.h"
#include "prog.h"

namespace Window
{
	static int win_w, win_h, win_w_half, win_h_half;
	static float win_w_div_h;
	static int displ_w, displ_h;
	static bool winfocus = 1;

	#if MANIFEST_WINDOW_RESIZABLE == 1
	static int saved_w, saved_h;
	#else
	static constexpr int saved_w = MANIFEST_WINDOW_W, saved_h = MANIFEST_WINDOW_H;
	#endif

	static SDL_Window *winhandle = 0;

	static const int display_num = 0;

	SDL_Window *Handle()
    {
    	return winhandle;
    }

	int W()        {return win_w;}
	int H()        {return win_h;}
	int W_m2()     {return win_w_half*2;}
	int H_m2()     {return win_h_half*2;}
	float Ratio()  {return win_w_div_h;}
	int W_half()   {return win_w_half;}
	int H_half()   {return win_h_half;}

	int DisplayNum() {return display_num;}
	int DisplayW()   {return displ_w;}
	int DisplayH()   {return displ_h;}

	static void CheckSize()
	{
		int w, h;
		SDL_GetWindowSize(winhandle, &w, &h);
		if (w == win_w && h == win_h)
			return;
		win_w = w;
		win_h = h;
		win_w_half = (win_w + 1) / 2;
		win_h_half = (win_h + 1) / 2;
		win_w_div_h = float(w) / float(h);
		Prog::Resize();
	}

	bool Focus()
	{
		return winfocus;
	}
}

namespace Input
{
	static bool relative_mouse = 0;
	int *board, *board_end, boardlen;
	static bool focus = 0;

	struct MouseData
    {
    	int x, y, rx, ry, mwx, mwy;
    	signed char buttons[32];
    	MouseData() : x(0), y(0), rx(0), ry(0), mwx(0), mwy(0), buttons{} {}
    };

    static MouseData anymouse;
    static std::map<uint32_t, MouseData> mousemap;

    int GetKeyState(unsigned int obj)
    {
    	if (int(obj) >= boardlen)
			return 0;
		else
			return board[obj];
    }

	bool WhatKeyIsPressed(unsigned int *p)
    {
    	for (int *it = board; it != board_end; it++)
		{
			if (*it == 1)
			{
				if (p)
					*p = it - board;
				return 1;
			}
		}
		return 0;
    }

    bool MouseFocus()
	{
		return focus;
	}


	int MouseX()
	{
		return anymouse.x;
	}
	int MouseY()
	{
		return anymouse.y;
	}
	int MouseRelX()
	{
		return anymouse.rx;
	}
	int MouseRelY()
	{
		return anymouse.ry;
	}
	int MouseWheelX()
	{
		return anymouse.mwx;
	}
	int MouseWheelY()
	{
		return anymouse.mwy;
	}
	int GetMouseButtonState(unsigned int n) // n: 0..31
	{
		return anymouse.buttons[n];
	}

	int MouseX_ID(uint32_t id)
	{
		return mousemap[id].x;
	}
	int MouseY_ID(uint32_t id)
	{
		return mousemap[id].y;
	}
	int MouseRelX_ID(uint32_t id)
	{
		return mousemap[id].rx;
	}
	int MouseRelY_ID(uint32_t id)
	{
		return mousemap[id].ry;
	}
	int MouseWheelX_ID(uint32_t id)
	{
		return mousemap[id].mwx;
	}
	int MouseWheelY_ID(uint32_t id)
	{
		return mousemap[id].mwy;
	}
	int GetMouseButtonState_ID(uint32_t id, unsigned int n) // n: 0..31
	{
		return mousemap[id].buttons[n];
	}

	bool WhatMouseButtonIsPressed(unsigned int *p)
	{
		for (int i = 0; i < 32; i++)
		{
			if (anymouse.buttons[i] == 1)
			{
				if (p)
					*p = i;
				return 1;
			}
		}
		return 0;
	}
	bool WhatMouseButtonIsPressed_ID(uint32_t id, unsigned int *p)
	{
		auto &ref = mousemap[id];
		for (int i = 0; i < 32; i++)
		{
			if (ref.buttons[i] == 1)
			{
				if (p)
					*p = i;
				return 1;
			}
		}
		return 0;
	}
	bool WhatMouseButtonAndIDIsPressed(unsigned int *p, uint32_t *id_p)
	{
		for (auto &it : mousemap)
		{
			for (int i = 0; i < 32; i++)
			{
				if (it.second.buttons[i] == 1)
				{
					if (p)
						*p = i;
					if (id_p)
						*id_p = it.first;
					return 1;
				}
			}
		}
		return 0;
	}
	bool WhatMouseIDHasButtonPressed(unsigned int n, uint32_t *p)
	{
		for (auto &it : mousemap)
		{
			if (it.second.buttons[n] == 1)
			{
				if (p)
					*p = it.first;
				return 1;
			}
		}
		return 0;
	}

	static void Init()
	{
		SDL_GetKeyboardState(&boardlen);
		board = new int[boardlen]; // Warning: no `delete` for this.
		std::memset(board, 0, boardlen * sizeof (int));
		board_end = board + boardlen;
	}

	namespace MouseMode
	{
		void Show()
		{
			SDL_ShowCursor(1);
		}
		void Hide()
		{
			SDL_ShowCursor(0);
		}

		void Release()
		{
			SDL_SetWindowGrab(Window::Handle(), (SDL_bool)0);
		}
		void Grab()
		{
			SDL_SetWindowGrab(Window::Handle(), (SDL_bool)1);
		}

		void NonRelative() // Also shows cursor (independently from Show()/Hide())
		{
			SDL_SetRelativeMouseMode((SDL_bool)0);
			relative_mouse = 0;
		}
		void Relative() // Also hides cursor (independently from Show()/Hide())
		{
			SDL_SetRelativeMouseMode((SDL_bool)1);
			relative_mouse = 1;
		}
	}
}

namespace System
{
	static constexpr double target_fps = MANIFEST_SYSTEM_FPS;
	static constexpr long double target_delta = 1.0l / target_fps;
	static uint64_t target_clockunit_delta;

	static double fps;
	static long double delta;
	static uint64_t curticktimepos;

	long double Delta()
	{
		return delta;
	}

	double FPS()
	{
		return fps;
	}

	static inline void FullScreen_Off()
	{
		SDL_SetWindowFullscreen(Window::Handle(), 0);
		#if MANIFEST_WINDOW_RESIZABLE == 1
		SDL_SetWindowSize(Window::Handle(), Window::saved_w, Window::saved_h);
		#endif
	}

	static inline void FullScreen_On()
	{
		#if MANIFEST_WINDOW_RESIZABLE == 1
		SDL_GetWindowSize(Window::Handle(), &Window::saved_w, &Window::saved_h);
		SDL_SetWindowSize(Window::Handle(), Window::DisplayW(), Window::DisplayH());
		#endif
		SDL_SetWindowFullscreen(Window::Handle(), SDL_WINDOW_FULLSCREEN);
	}

	static inline void FullScreenHotkeyCheck()
	{
		#if MANIFEST_SYSTEM_ENABLE_ALT_ENTER == 1
		static bool fullscr = MANIFEST_WINDOW_INITIAL_STATE == 2;
		if (Input::KeyPressed(Input::Key_Enter()) && (Input::KeyDown(Input::Key_Alt_L()) || Input::KeyDown(Input::Key_Alt_R())))
		{
			if (fullscr)
			{
				fullscr = 0;
				FullScreen_Off();
			}
			else
			{
				fullscr = 1;
				FullScreen_On();
			}
		}
		#endif
	}

	static inline void EventsTick()
	{
		for (int *it = Input::board; it != Input::board_end; it++)
		{
			switch (*it)
			{
			  case 0:           break;
			  case 1:  *it = 2; break;
			  case 2:           break;
			  default: *it = 0; break;
			}
		}

		Input::anymouse.rx = Input::anymouse.ry = Input::anymouse.mwx = Input::anymouse.mwy = 0;
		for (int i = 0; i < 32; i++)
		{
			switch (Input::anymouse.buttons[i])
			{
			  case 0:                                  break;
			  case 1:  Input::anymouse.buttons[i] = 2; break;
			  case 2:                                  break;
			  default: Input::anymouse.buttons[i] = 0; break;
			}
		}

		for (auto &it : Input::mousemap)
		{
			it.second.rx = it.second.ry = it.second.mwx = it.second.mwy = 0;
			for (int i = 0; i < 32; i++)
			{
				switch (it.second.buttons[i])
				{
				  case 0:                            break;
				  case 1:  it.second.buttons[i] = 2; break;
				  case 2:                            break;
				  default: it.second.buttons[i] = 0; break;
				}
			}
		}

		SDL_Event curevent;
		while (SDL_PollEvent(&curevent))
		{
			switch (curevent.type)
			{
			  case SDL_QUIT:
				Main::Exit();
				break;
			  case SDL_MOUSEMOTION:
				{
					Input::MouseData &ref = Input::mousemap[curevent.motion.which];
					ref.x = Input::anymouse.x = curevent.motion.x;
					ref.y = Input::anymouse.y = curevent.motion.y;
					if (Window::winfocus)
					{
						ref.rx += curevent.motion.xrel;
						ref.ry += curevent.motion.yrel;
						#if MANIFEST_FEATURE_TOUCH_INPUT == 0
						Input::anymouse.rx += curevent.motion.xrel;
						Input::anymouse.ry += curevent.motion.yrel;
						#endif
					}
				}
				break;
			  case SDL_MOUSEWHEEL:
				{
					Input::MouseData &ref = Input::mousemap[curevent.wheel.which];
					Input::anymouse.mwx += curevent.wheel.x;
					Input::anymouse.mwy += curevent.wheel.y;
					ref.mwx += curevent.wheel.x;
					ref.mwy += curevent.wheel.y;
				}
				break;
			  case SDL_KEYDOWN:
				if (curevent.key.repeat)
					break;
				Input::board[curevent.key.keysym.scancode] = 1;
				break;
			  case SDL_KEYUP:
				if (curevent.key.repeat)
					break;
				Input::board[curevent.key.keysym.scancode] = 3;
				break;
			  case SDL_MOUSEBUTTONDOWN:
				Input::mousemap[curevent.button.which].buttons[curevent.button.button-1] = Input::anymouse.buttons[curevent.button.button-1] = 1;
				break;
			  case SDL_MOUSEBUTTONUP:
				Input::mousemap[curevent.button.which].buttons[curevent.button.button-1] = Input::anymouse.buttons[curevent.button.button-1] = 3;
				break;
			  case SDL_WINDOWEVENT:
			  	switch (curevent.window.event)
				{
				  case SDL_WINDOWEVENT_FOCUS_LOST:
					if (Input::relative_mouse)
						SDL_SetRelativeMouseMode((SDL_bool)0);
					Window::winfocus = 0;
					break;
				  case SDL_WINDOWEVENT_FOCUS_GAINED:
				  	if (Input::relative_mouse)
						SDL_SetRelativeMouseMode((SDL_bool)1);
					Window::winfocus = 1;
					break;
				}
				break;
			}
		}

		#if MANIFEST_FEATURE_TOUCH_INPUT == 1
		for (auto &it : Input::mousemap)
		{
			if (it.second.buttons[0] == 1)
				it.second.rx = it.second.ry = 0;
			Input::anymouse.rx += it.second.rx;
			Input::anymouse.ry += it.second.ry;
		}
		#endif

		Input::focus = SDL_GetMouseFocus();
	}

	void Init()
	{
		EXEC_ONCE;

		static constexpr uint32_t sdlwin_flags = SDL_WINDOW_OPENGL
											   | (MANIFEST_WINDOW_RESIZABLE          ? SDL_WINDOW_RESIZABLE  : 0)
											   | (MANIFEST_WINDOW_INITIAL_STATE == 1 ? SDL_WINDOW_MAXIMIZED  : 0)
											   | (MANIFEST_WINDOW_INITIAL_STATE == 2 ? SDL_WINDOW_FULLSCREEN : 0);
		SDL_SetHint(SDL_HINT_RENDER_VSYNC, MANIFEST_SYSTEM_VSYNC ? "1" : "0");
		SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 8);
		SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 8);
		SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 8);
		SDL_GL_SetAttribute(SDL_GL_ALPHA_SIZE, 8);
		SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, MANIFEST_OPENGL_MAJOR);
		SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, MANIFEST_OPENGL_MINOR);
		#if MANIFEST_OPENGL_MAJOR >= 3
		int profiles[3] {SDL_GL_CONTEXT_PROFILE_CORE, SDL_GL_CONTEXT_PROFILE_COMPATIBILITY, SDL_GL_CONTEXT_PROFILE_ES};
		SDL_GL_SetAttribute(SDL_GL_CONTEXT_PROFILE_MASK, profiles[MANIFEST_OPENGL_PROFILE]);
		SDL_GL_SetAttribute(SDL_GL_CONTEXT_FLAGS, MANIFEST_OPENGL_FORWARD_COMPAT ? SDL_GL_CONTEXT_FORWARD_COMPATIBLE_FLAG : 0);
		#endif

		SDL_DisplayMode displ_data;
		SDL_GetDesktopDisplayMode(Window::DisplayNum(), &displ_data);
		Window::displ_w = displ_data.w;
		Window::displ_h = displ_data.h;

		if (!(Window::winhandle = SDL_CreateWindow(MANIFEST_APPNAME, SDL_WINDOWPOS_CENTERED_DISPLAY(Window::DisplayNum()), SDL_WINDOWPOS_CENTERED_DISPLAY(Window::DisplayNum()),
												   #if MANIFEST_WINDOW_INITIAL_STATE == 2 && MANIFEST_WINDOW_RESIZABLE == 1
												   Window::displ_w, Window::displ_h,
												   #else
												   MANIFEST_WINDOW_W, MANIFEST_WINDOW_H,
												   #endif
												   sdlwin_flags)))
			Main::Exit(Jo("Window creation failed: `", SDL_GetError(), '`'));

		#if MANIFEST_WINDOW_RESIZABLE == 1
		#if MANIFEST_WINDOW_MIN_SIZE == 1
		SDL_SetWindowMinimumSize(Window::winhandle, MANIFEST_WINDOW_MINW, MANIFEST_WINDOW_MINH);
		#endif
		#if MANIFEST_WINDOW_MAX_SIZE == 1
		SDL_SetWindowMinimumSize(Window::winhandle, MANIFEST_WINDOW_MAXW, MANIFEST_WINDOW_MAXH);
		#endif
		#endif

		#if MANIFEST_WINDOW_RESIZABLE == 1 && MANIFEST_WINDOW_INITIAL_STATE != 0
		SDL_GetWindowSize(Window::winhandle, &Window::win_w, &Window::win_h);
		#else
		Window::win_w = MANIFEST_WINDOW_W;
		Window::win_h = MANIFEST_WINDOW_H;
		#endif
		Window::win_w_half = (Window::win_w + 1) / 2;
		Window::win_h_half = (Window::win_h + 1) / 2;
		Window::win_w_div_h = float(Window::win_w) / float(Window::win_h);

		#if MANIFEST_WINDOW_INITIAL_STATE == 2 && MANIFEST_WINDOW_RESIZABLE == 1
		SDL_GetWindowSize(Window::winhandle, &Window::saved_w, &Window::saved_h);
		#endif

		target_clockunit_delta = Utils::SecondsToClockUnits(target_delta);

		Input::Init();

		GFX::Init();
	}

	void PreTick()
	{
		curticktimepos = Utils::Clock();

		Window::CheckSize();

		GFX::PreTick();

		EventsTick();

		FullScreenHotkeyCheck();
	}



	void PostTick()
	{
		GFX::PostTick();

		Audio::Tick();

		uint64_t ticklen = Utils::Clock() - curticktimepos;
		fps = 1.0 / delta;
		if (ticklen < target_clockunit_delta)
		{
			delta = target_delta;
			Utils::WaitClockUnits(target_clockunit_delta - ticklen);
		}
		else
		{
			delta = Utils::ClockUnitsToSecs(ticklen);
		}
	}
}